"""Tensor product operation."""
import numpy as np


def tensor(*args) -> np.ndarray:
    r"""
    Compute the Kronecker tensor product [WikTensor]_.

    Tensor two matrices or vectors together using the standard Kronecker
    operation provided from numpy.

    Given two matrices :math:`A` and :math:`B`, computes :math:`A \otimes B`.
    The same concept also applies to two vectors :math:`v` and :math:`w` which
    computes :math:`v \otimes w`.

    One may also compute the tensor product one matrix `n` times with itself.

    For a matrix, :math:`A` and an integer :math:`n`, the result of this
    function computes :math:`A^{\otimes n}`.

    Similarly for a vector :math:`v` and an integer :math:`n`, the result of
    of this function computes :math:`v^{\otimes n}`.

    One may also perform the tensor product on a list of matrices.

    Given a list of :math:`n` matrices :math:`A_1, A_2, \ldots, A_n` the result
    of this function computes

    .. math::
        A_1 \otimes A_2 \otimes \ldots \otimes A_n.

    Similarly, for a list of :math:`n` vectors :math:`v_1, v_2, \ldots, v_n`,
    the result of this function computes

    .. math::
        v_1 \otimes v_2 \otimes \ldots \otimes v_n.

    Examples
    ==========

    Tensor product two matrices or vectors

    Consider the following ket vector

    .. math::
        e_0 = \left[1, 0 \right]^{\text{T}}.

    Computing the following tensor product

    .. math:
        e_0 \otimes e_0 = \[1, 0, 0, 0 \]^{\text{T}}.

    This can be accomplished in :code:`toqito` as follows.

    >>> from toqito.states import basis
    >>> from toqito.matrix_ops import tensor
    >>> e_0 = basis(2, 0)
    >>> tensor(e_0, e_0)
    [[1],
     [0],
     [0],
     [0]]

    Tensor product one matrix :math:`n` times with itself.

    We may also tensor some element with itself some integer number of times.
    For instance we can compute

    .. math::
        e_0^{\otimes 3} = \left[1, 0, 0, 0, 0, 0, 0, 0 \right]^{\text{T}}

    in `toqito` as follows.

    >>> from toqito.states import basis
    >>> from toqito.matrix_ops import tensor
    >>> e_0 = basis(2, 0)
    >>> tensor(e_0, 3)
    [[1],
     [0],
     [0],
     [0],
     [0],
     [0],
     [0],
     [0]]

    Perform the tensor product on a list of vectors or matrices.

    If we wish to compute the tensor product against more than two matrices or
    vectors, we can feed them in as a `list`. For instance, if we wish to
    compute :math:`e_0 \otimes e_1 \otimes e_0`, we can do
    so as follows.

    >>> from toqito.states import basis
    >>> from toqito.matrix_ops import tensor
    >>> e_0, e_1 = basis(2, 0), basis(2, 1)
    >>> tensor([e_0, e_1, e_0])
    [[0],
     [0],
     [1],
     [0],
     [0],
     [0],
     [0],
     [0]]

    References
    ==========
    .. [WikTensor] Wikipedia: Tensor product
        https://en.wikipedia.org/wiki/Tensor_product

    :param args: Input to the tensor function is expected to be either:
        - List[np.ndarray]: List of numpy matrices,
        - np.ndarray, ... , np.ndarray: An arbitrary number of numpy arrays,
        - np.ndarray, int: A numpy array and an integer.
    :return: The computed tensor product.
    """
    result = None

    # Input is provided as a list of numpy matrices.
    if len(args) == 1 and isinstance(args[0], list):
        if len(args[0]) == 1:
            return args[0][0]
        if len(args[0]) == 2:
            return np.kron(args[0][0], args[0][1])
        if len(args[0]) >= 3:
            result = args[0][0]
            for i in range(1, len(args[0])):
                result = np.kron(result, args[0][i])
        return result

    if len(args) == 1 and isinstance(args[0], np.ndarray):
        # If the numpy array is just a single matrix, so the dimensions are
        # provided as an (x, y)-tuple.
        if len(args[0].shape) == 2:
            return args[0]
        if len(args[0]) == 2:
            return np.kron(args[0][0], args[0][1])
        if len(args[0]) >= 3:
            result = args[0][0]
            for i in range(1, len(args[0])):
                result = np.kron(result, args[0][i])
        return result

    # Tensor product one matrix `n` times with itself.
    if len(args) == 2 and isinstance(args[1], int):
        num_tensor = args[1]
        if num_tensor == 1:
            return args[0]
        if num_tensor == 2:
            return np.kron(args[0], args[0])
        if num_tensor >= 3:
            result = np.kron(args[0], args[0])
            for _ in range(2, num_tensor):
                result = np.kron(result, args[0])
        return result

    # Tensor product between two or more matrices.
    if len(args) == 2:
        return np.kron(args[0], args[1])
    if len(args) >= 3:
        result = args[0]
        for i in range(1, len(args)):
            result = np.kron(result, args[i])
        return result
    raise ValueError("The `tensor` function must take either a matrix or vector.")



"""Test is_positive_semidefinite."""
import numpy as np

from toqito.matrix_props import is_positive_semidefinite


def test_is_positive_semidefinite():
    """Test that positive semidefinite matrix returns True."""
    mat = np.array([[1, -1], [-1, 1]])
    np.testing.assert_equal(is_positive_semidefinite(mat), True)


def test_is_not_positive_semidefinite():
    """Test that non-positive semidefinite matrix returns False."""
    mat = np.array([[-1, -1], [-1, -1]])
    np.testing.assert_equal(is_positive_semidefinite(mat), False)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Test is_square."""
import numpy as np

from toqito.matrix_props import is_square


def test_is_square():
    """Test that square matrix returns True."""
    mat = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    np.testing.assert_equal(is_square(mat), True)


def test_is_not_square():
    """Test that non-square matrix returns False."""
    mat = np.array([[1, 2, 3], [4, 5, 6]])
    np.testing.assert_equal(is_square(mat), False)


def test_is_square_invalid():
    """Input must be a matrix."""
    with np.testing.assert_raises(ValueError):
        is_square(np.array([-1, 1]))


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Test is_mixed."""
import numpy as np

from toqito.state_props import is_mixed
from toqito.states import basis


def test_is_mixed():
    """Return True for mixed quantum state."""
    e_0, e_1 = basis(2, 0), basis(2, 1)
    rho = 3 / 4 * e_0 * e_0.conj().T + 1 / 4 * e_1 * e_1.conj().T
    np.testing.assert_equal(is_mixed(rho), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Kronecker product for CVXPY objects."""
from typing import Union

import numpy as np
import cvxpy
from cvxpy.expressions.expression import Expression


def cvx_kron(
    expr_1: Union[np.ndarray, Expression], expr_2: Union[np.ndarray, Expression]
) -> Expression:
    """
    Compute Kronecker product between CVXPY objects.

    By default, CVXPY does not support taking the Kronecker product when the argument on the left is
    equal to a CVXPY object and the object on the right is equal to a numpy object.

    At most one of :code:`expr_1` and :code:`b` may be CVXPY Variable objects.

    Kudos to Riley J. Murray for this function:
    https://github.com/cvxgrp/cvxpy/issues/457

    :param expr_1: 2D numpy ndarray, or a CVXPY Variable with expr_1.ndim == 2
    :param expr_2: 2D numpy ndarray, or a CVXPY Variable with expr_2.ndim == 2
    :return: The tensor product of two CVXPY expressions.
    """
    expr = np.kron(expr_1, expr_2)
    num_rows = expr.shape[0]
    rows = [cvxpy.hstack(expr[i, :]) for i in range(num_rows)]
    full_expr = cvxpy.vstack(rows)
    return full_expr



"""CNOT matrix."""
import numpy as np


def cnot() -> np.ndarray:
    r"""
    Produce the CNOT matrix [WikCNOT]_.

    The CNOT matrix is defined as

    .. math::
        \text{CNOT} =
        \begin{pmatrix}
            1 & 0 & 0 & 0 \\
            0 & 1 & 0 & 0 \\
            0 & 0 & 0 & 1 \\
            0 & 0 & 1 & 0
        \end{pmatrix}.

    Examples
    ==========

    >>> from toqito.matrices import cnot
    >>> cnot()
    [[1 0 0 0]
     [0 1 0 0]
     [0 0 0 1]
     [0 0 1 0]]

    References
    ==========
    .. [WikCNOT] Wikipedia: Controlled NOT gate
        https://en.wikipedia.org/wiki/Controlled_NOT_gate

    :return: The CNOT matrix.
    """
    return np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]])



"""Test random_povm."""
import numpy as np

from toqito.random import random_povm


def test_random_povm_unitary_not_real():
    """Generate random POVMs and check that they sum to the identity."""
    dim, num_inputs, num_outputs = 2, 2, 2
    povms = random_povm(dim, num_inputs, num_outputs)
    np.testing.assert_equal(
        np.allclose(povms[:, :, 0, 0] + povms[:, :, 0, 1], np.identity(dim)), True
    )


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Test negativity."""
import numpy as np

from toqito.state_props import negativity


def test_negativity_rho():
    """Test for negativity on rho."""
    test_input_mat = np.array(
        [[1 / 2, 0, 0, 1 / 2], [0, 0, 0, 0], [0, 0, 0, 0], [1 / 2, 0, 0, 1 / 2]]
    )
    np.testing.assert_equal(np.isclose(negativity(test_input_mat), 1 / 2), True)


def test_negativity_rho_dim_int():
    """Test for negativity on rho."""
    test_input_mat = np.array(
        [[1 / 2, 0, 0, 1 / 2], [0, 0, 0, 0], [0, 0, 0, 0], [1 / 2, 0, 0, 1 / 2]]
    )
    np.testing.assert_equal(np.isclose(negativity(test_input_mat, 2), 1 / 2), True)


def test_negativity_invalid_rho_dim_int():
    """Invalid dim parameters."""
    with np.testing.assert_raises(ValueError):
        test_input_mat = np.array(
            [[1 / 2, 0, 0, 1 / 2], [0, 0, 0, 0], [0, 0, 0, 0], [1 / 2, 0, 0, 1 / 2]]
        )
        negativity(test_input_mat, 5)


def test_negativity_invalid_rho_dim_vec():
    """Invalid dim parameters."""
    with np.testing.assert_raises(ValueError):
        test_input_mat = np.array(
            [[1 / 2, 0, 0, 1 / 2], [0, 0, 0, 0], [0, 0, 0, 0], [1 / 2, 0, 0, 1 / 2]]
        )
        negativity(test_input_mat, [2, 5])


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Test symmetric_projection."""
import numpy as np

from toqito.perms import symmetric_projection


def test_symmetric_projection_dim_2_pval_1():
    """Symmetric_projection where the dimension is 2 and p_val is 1."""
    res = symmetric_projection(dim=2, p_val=1)
    expected_res = np.array([[1, 0], [0, 1]])

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_symmetric_projection_2():
    """Generates the symmetric_projection where the dimension is 2."""
    res = symmetric_projection(dim=2)
    expected_res = np.array([[1, 0, 0, 0], [0, 0.5, 0.5, 0], [0, 0.5, 0.5, 0], [0, 0, 0, 1]])

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_symmetric_projection_dim_2_partial_true():
    """Symmetric_projection where the dimension is 2 and partial is True."""
    res = symmetric_projection(dim=2, p_val=2, partial=True)
    expected_res = symmetric_projection(dim=2, p_val=2, partial=False)

    bool_mat = np.isclose(res @ res.conj().T, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_symmetric_projection_dim_2_pval_3():
    """Symmetric_projection where the dimension is 2 and p_val is 3."""
    res = symmetric_projection(dim=2, p_val=3)
    expected_res = np.array(
        [
            [1, 0, 0, 0, 0, 0, 0, 0],
            [0, 1 / 3, 1 / 3, 0, 1 / 3, 0, 0, 0],
            [0, 1 / 3, 1 / 3, 0, 1 / 3, 0, 0, 0],
            [0, 0, 0, 1 / 3, 0, 1 / 3, 1 / 3, 0],
            [0, 1 / 3, 1 / 3, 0, 1 / 3, 0, 0, 0],
            [0, 0, 0, 1 / 3, 0, 1 / 3, 1 / 3, 0],
            [0, 0, 0, 1 / 3, 0, 1 / 3, 1 / 3, 0],
            [0, 0, 0, 0, 0, 0, 0, 1],
        ]
    )

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_symmetric_projection_dim_3_pval_2():
    """Symmetric_projection where the dimension is 3 and p_val is 2."""
    res = symmetric_projection(dim=3, p_val=2)
    expected_res = np.array(
        [
            [1, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 1 / 2, 0, 1 / 2, 0, 0, 0, 0, 0],
            [0, 0, 1 / 2, 0, 0, 0, 1 / 2, 0, 0],
            [0, 1 / 2, 0, 1 / 2, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 1 / 2, 0, 1 / 2, 0],
            [0, 0, 1 / 2, 0, 0, 0, 1 / 2, 0, 0],
            [0, 0, 0, 0, 0, 1 / 2, 0, 1 / 2, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 1],
        ]
    )

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_symmetric_projection_dim_4_pval_2():
    """Symmetric_projection where the dimension is 4 and p_val is 2."""
    res = symmetric_projection(dim=4, p_val=2)
    expected_res = np.array(
        [
            [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 1 / 2, 0, 0, 1 / 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 1 / 2, 0, 0, 0, 0, 0, 1 / 2, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 1 / 2, 0, 0, 0, 0, 0, 0, 0, 0, 1 / 2, 0, 0, 0],
            [0, 1 / 2, 0, 0, 1 / 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 1 / 2, 0, 0, 1 / 2, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 1 / 2, 0, 0, 0, 0, 0, 1 / 2, 0, 0],
            [0, 0, 1 / 2, 0, 0, 0, 0, 0, 1 / 2, 0, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 1 / 2, 0, 0, 1 / 2, 0, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1 / 2, 0, 0, 1 / 2, 0],
            [0, 0, 0, 1 / 2, 0, 0, 0, 0, 0, 0, 0, 0, 1 / 2, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 1 / 2, 0, 0, 0, 0, 0, 1 / 2, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1 / 2, 0, 0, 1 / 2, 0],
            [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1],
        ]
    )

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_symmetric_projection_dim_4_pval_2_partial_true():
    """Dimension is 4, p_val is 2, and partial is True."""
    res = symmetric_projection(dim=4, p_val=2, partial=True)
    expected_res = symmetric_projection(dim=4, p_val=2, partial=False)

    bool_mat = np.isclose(res @ res.conj().T, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Brauer states."""
import numpy as np

from toqito.matrix_ops import tensor
from toqito.perms import permute_systems, perfect_matchings
from toqito.states import max_entangled


def brauer(dim: int, p_val: int) -> np.ndarray:
    r"""
    Produce all Brauer states [WikBrauer]_.

    Produce a matrix whose columns are all of the (unnormalized) "Brauer" states: states that are
    the :code:`p_val`-fold tensor product of the standard maximally-entangled pure state on
    :code:`dim` local dimensions. There are many such states, since there are many different ways to
    group the :code:`2 * p_val` parties into :code:`p_val` pairs (with each pair corresponding to
    one maximally-entangled state).

    The exact number of such states is:

    ```python
    np.factorial(2 * p_val) / (np.factorial(p_val) * 2**p_val)
    ```

    which is the number of columns of the returned matrix.

    This function has been adapted from QETLAB.

    Examples
    ==========

    Generate a matrix whose columns are all Brauer states on 4 qubits.

    >>> from toqito.states import brauer
    >>> brauer(2, 2)
    [[1. 1. 1.]
     [0. 0. 0.]
     [0. 0. 0.]
     [1. 0. 0.]
     [0. 0. 0.]
     [0. 1. 0.]
     [0. 0. 1.]
     [0. 0. 0.]
     [0. 0. 0.]
     [0. 0. 1.]
     [0. 1. 0.]
     [0. 0. 0.]
     [1. 0. 0.]
     [0. 0. 0.]
     [0. 0. 0.]
     [1. 1. 1.]]

    References
    ==========
    .. [WikBrauer] Wikipedia: Brauer algebra
        https://en.wikipedia.org/wiki/Brauer_algebra

    :param dim: Dimension of each local subsystem
    :param p_val: Half of the number of parties (i.e., the state that this function computes will
                  live in :math:`(\mathbb{C}^D)^{\otimes 2 P})`
    :return: Matrix whose columns are all of the unnormalized Brauer states.
    """
    # The Brauer states are computed from perfect matchings of the complete graph. So compute all
    # perfect matchings first.
    phi = tensor(max_entangled(dim, False, False), p_val)
    matchings = perfect_matchings(2 * p_val)
    num_matchings = matchings.shape[0]
    state = np.zeros((dim ** (2 * p_val), num_matchings))

    # Turn these perfect matchings into the corresponding states.
    for i in range(num_matchings):
        state[:, i] = permute_systems(
            phi, matchings[i, :], dim * np.ones((1, 2 * p_val), dtype=int)[0]
        )
    return state



"""Calculate the Choi rank of a channel."""
from typing import List, Union

import numpy as np
from toqito.channel_ops import kraus_to_choi


def choi_rank(phi: Union[np.ndarray, List[List[np.ndarray]]]) -> int:
    r"""
    Calculate the rank of the Choi representation of a quantum channel [WatChoiRank18]_.

    Examples
    ==========

    The transpose map can be written either in Choi representation (as a
    SWAP operator) or in Kraus representation. If we choose the latter, it
    will be given by the following matrices:

    .. math::
        \begin{equation}
            \begin{aligned}
                \frac{1}{\sqrt{2}}
                \begin{pmatrix}
                    0 & i \\ -i & 0
                \end{pmatrix}, &\quad
                \frac{1}{\sqrt{2}}
                \begin{pmatrix}
                    0 & 1 \\
                    1 & 0
                \end{pmatrix}, \\
                \begin{pmatrix}
                    1 & 0 \\
                    0 & 0
                \end{pmatrix}, &\quad
                \begin{pmatrix}
                    0 & 0 \\
                    0 & 1
                \end{pmatrix}.
            \end{aligned}
        \end{equation}

    and can be generated in :code:`toqito` with the following list:

    >>> import numpy as np
    >>> kraus_1 = np.array([[1, 0], [0, 0]])
    >>> kraus_2 = np.array([[1, 0], [0, 0]]).conj().T
    >>> kraus_3 = np.array([[0, 1], [0, 0]])
    >>> kraus_4 = np.array([[0, 1], [0, 0]]).conj().T
    >>> kraus_5 = np.array([[0, 0], [1, 0]])
    >>> kraus_6 = np.array([[0, 0], [1, 0]]).conj().T
    >>> kraus_7 = np.array([[0, 0], [0, 1]])
    >>> kraus_8 = np.array([[0, 0], [0, 1]]).conj().T
    >>> kraus_ops = [
    >>>     [kraus_1, kraus_2],
    >>>     [kraus_3, kraus_4],
    >>>     [kraus_5, kraus_6],
    >>>     [kraus_7, kraus_8],
    >>> ]

    To calculate its Choi rank, we proceed in the following way:

    >>> from toqito.channel_props import choi_rank
    >>> choi_rank(kraus_ops)
    4

    We can the verify the associated Choi representation (the SWAP gate)
    gets the same Choi rank:

    >>> choi_matrix = np.array([[1,0,0,0],[0,0,1,0],[0,1,0,0],[0,0,0,1]])
    >>> choi_rank(choi_matrix)
    4

    References
    ==========

    .. [WatChoiRank18] Watrous, John.
        "The Theory of Quantum Information."
        Section: "2.2 Quantum Channels".
        Cambridge University Press, 2018.

    :param phi: Either a Choi matrix or a list of Kraus operators
    :return: The Choi rank of the provided channel representation.
    """
    if isinstance(phi, list):
        phi = kraus_to_choi(phi)
    elif not isinstance(phi, np.ndarray):
        raise ValueError("Not a valid Choi matrix.")

    return np.linalg.matrix_rank(phi)



"""Test gen_bell."""
import numpy as np

from toqito.states import bell
from toqito.states import gen_bell


def test_gen_bell_0_0_2():
    """Generalized Bell state for k_1 = k_2 = 0 and dim = 2."""
    dim = 2
    k_1 = 0
    k_2 = 0

    expected_res = bell(0) * bell(0).conj().T

    res = gen_bell(k_1, k_2, dim)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_gen_bell_0_1_2():
    """Generalized Bell state for k_1 = 0, k_2 = 1 and dim = 2."""
    dim = 2
    k_1 = 0
    k_2 = 1

    expected_res = bell(1) * bell(1).conj().T

    res = gen_bell(k_1, k_2, dim)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_gen_bell_1_0_2():
    """Generalized Bell state for k_1 = 1, k_2 = 0 and dim = 2."""
    dim = 2
    k_1 = 1
    k_2 = 0

    expected_res = bell(2) * bell(2).conj().T

    res = gen_bell(k_1, k_2, dim)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_gen_bell_1_1_2():
    """Generalized Bell state for k_1 = 1, k_2 = 1 and dim = 2."""
    dim = 2
    k_1 = 1
    k_2 = 1

    expected_res = bell(3) * bell(3).conj().T

    res = gen_bell(k_1, k_2, dim)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Unvec operation."""
from typing import List, Optional

import numpy as np


def unvec(vector: np.ndarray, shape: Optional[List[int]] = None) -> np.ndarray:
    r"""
    Perform the unvec operation on a vector to obtain a matrix [Rigetti2020]_.

    Takes a column vector and transforms it into a :code:`shape[0]`-by-:code`shape[1]` matrix.
    This operation is the inverse of :code:`vec` operation in :code:`toqito`.

    For instance, for the following column vector

    .. math::
        u = \begin{pmatrix} 1 \\ 3 \\ 2 \\ 4 \end{pmatrix},

    it holds that

    .. math::
        \text{unvec}(u) =
        begin{pmatrix}
            1 & 2 \\
            3 & 4
        \end{pmatrix}

    More formally, the vec operation is defined by

    .. math::
        \text{unvec}(e_a \otimes e_b) = E_{a,b}

    for all :math:`a` and :math:`b` where

    .. math::
        E_{a,b}(c,d) = \begin{cases}
                          1 & \text{if} \ (c,d) = (a,b) \\
                          0 & \text{otherwise}
                        \end{cases}

    for all :math:`c` and :math:`d` and where

    .. math::
        e_a(b) = \begin{cases}
                     1 & \text{if} \ a = b \\
                     0 & \text{if} \ a \not= b
                 \end{cases}

    for all :math:`a` and :math:`b`.

    This function has been adapted from [Rigetti20]_.

    Examples
    ==========

    Consider the following vector

    .. math::
        u = \begin{pmatrix} 1 \\ 3 \\ 2 \\ 4 \end{pmatrix}

    Performing the :math:`\text{unvec}` operation on :math:`u` yields

    .. math::
        \text{unvec}(u) = \begin{pmatrix} 1 & 2 \\ 3 & 4 \end{pmatrix}

    >>> from toqito.matrix_ops import unvec
    >>> import numpy as np
    >>> u = np.array([1, 2, 3, 4])
    >>> unvec(u)
    [[1 2]
     [3 4]]

    See Also
    ========
    vec

    References
    ==========
    .. [Rigetti2020] Forest Benchmarking (Rigetti).
        https://github.com/rigetti/forest-benchmarking

    :param vector: A (:code:`shape[0] * shape[1]`)-by-1 numpy array.
    :param shape: The shape of the output matrix; by default, the matrix is assumed to be square.
    :return: Returns a :code:`shape[0]`-by-:code:`shape[1]` matrix.
    """
    vector = np.asarray(vector)
    if shape is None:
        dim = int(np.sqrt(vector.size))
        shape = dim, dim
    mat = vector.reshape(*shape).T
    return mat



"""Test gell_mann."""
import numpy as np

from scipy.sparse import csr_matrix

from toqito.matrices import gell_mann


def test_gell_mann_idx_0():
    """Gell-Mann operator for index = 0."""
    expected_res = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
    res = gell_mann(0)

    bool_mat = np.isclose(expected_res, res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_gell_mann_idx_1():
    """Gell-Mann operator for index = 1."""
    expected_res = np.array([[0, 1, 0], [1, 0, 0], [0, 0, 0]])
    res = gell_mann(1)

    bool_mat = np.isclose(expected_res, res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_gell_mann_idx_2():
    """Gell-Mann operator for index = 2."""
    expected_res = np.array([[0, -1j, 0], [1j, 0, 0], [0, 0, 0]])
    res = gell_mann(2)

    bool_mat = np.isclose(expected_res, res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_gell_mann_idx_3():
    """Gell-Mann operator for index = 3."""
    expected_res = np.array([[1, 0, 0], [0, -1, 0], [0, 0, 0]])
    res = gell_mann(3)

    bool_mat = np.isclose(expected_res, res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_gell_mann_idx_4():
    """Gell-Mann operator for index = 4."""
    expected_res = np.array([[0, 0, 1], [0, 0, 0], [1, 0, 0]])

    res = gell_mann(4)

    bool_mat = np.isclose(expected_res, res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_gell_mann_idx_5():
    """Gell-Mann operator for index = 5."""
    expected_res = np.array([[0, 0, -1j], [0, 0, 0], [1j, 0, 0]])

    res = gell_mann(5)

    bool_mat = np.isclose(expected_res, res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_gell_mann_idx_6():
    """Gell-Mann operator for index = 6."""
    expected_res = np.array([[0, 0, 0], [0, 0, 1], [0, 1, 0]])

    res = gell_mann(6)

    bool_mat = np.isclose(expected_res, res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_gell_mann_idx_7():
    """Gell-Mann operator for index = 7."""
    expected_res = np.array([[0, 0, 0], [0, 0, -1j], [0, 1j, 0]])

    res = gell_mann(7)

    bool_mat = np.isclose(expected_res, res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_gell_mann_idx_8():
    """Gell-Mann operator for index = 8."""
    expected_res = np.array([[1, 0, 0], [0, 1, 0], [0, 0, -2]]) / np.sqrt(3)
    res = gell_mann(8)

    bool_mat = np.isclose(expected_res, res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_gell_mann_invalid_idx():
    """Invalid Gell-Mann parameters."""
    with np.testing.assert_raises(ValueError):
        gell_mann(9)


def test_gell_mann_sparse():
    """Test sparse Gell-Mann matrix."""
    res = gell_mann(3, is_sparse=True)
    np.testing.assert_equal(isinstance(res, csr_matrix), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Test perfect_matchings."""
import numpy as np

from toqito.perms import perfect_matchings


def test_perfect_matchings_num_4():
    """All perfect matchings of size 4."""
    res = perfect_matchings(4)
    expected_res = np.array([[1, 2, 3, 4], [1, 3, 2, 4], [1, 4, 3, 2]])
    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_perfect_matchings_num_4_list():
    """All perfect matchings of size 4 with input as list."""
    res = perfect_matchings([1, 2, 3, 4])
    expected_res = np.array([[1, 2, 3, 4], [1, 3, 2, 4], [1, 4, 3, 2]])
    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_perfect_matchings_odd():
    """There are no perfect matchings of an odd number of objects."""
    res = perfect_matchings(5)
    expected_res = np.zeros((0, 5))
    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_perfect_matchings_num_6():
    """All perfect matchings of size 6."""
    res = perfect_matchings(6)

    expected_res = np.array(
        [
            [1, 2, 3, 4, 5, 6],
            [1, 2, 3, 5, 4, 6],
            [1, 2, 3, 6, 5, 4],
            [1, 3, 2, 4, 5, 6],
            [1, 3, 2, 5, 4, 6],
            [1, 3, 2, 6, 5, 4],
            [1, 4, 3, 2, 5, 6],
            [1, 4, 3, 5, 2, 6],
            [1, 4, 3, 6, 5, 2],
            [1, 5, 3, 4, 2, 6],
            [1, 5, 3, 2, 4, 6],
            [1, 5, 3, 6, 2, 4],
            [1, 6, 3, 4, 5, 2],
            [1, 6, 3, 5, 4, 2],
            [1, 6, 3, 2, 5, 4],
        ]
    )
    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Schmidt decomposition operation."""
from typing import List, Tuple, Union
from scipy.sparse import issparse, linalg

import numpy as np


def schmidt_decomposition(
    vec: np.ndarray, dim: Union[int, List[int], np.ndarray] = None, k_param: int = 0
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    r"""
    Compute the Schmidt decomposition of a bipartite vector [WikSD]_.

    Examples
    ==========

    Consider the :math:`3`-dimensional maximally entangled state

    .. math::
        u = \frac{1}{\sqrt{3}} \left( |000 \rangle + |111 \rangle + |222 \rangle \right)

    We can generate this state using the :code:`toqito` module as follows.

    >>> from toqito.states import max_entangled
    >>> max_entangled(3)
    [[0.57735027],
     [0.        ],
     [0.        ],
     [0.        ],
     [0.57735027],
     [0.        ],
     [0.        ],
     [0.        ],
     [0.57735027]]

    Computing the Schmidt decomposition of :math:`u`, we can obtain the corresponding singular
    values of :math:`u` as

    .. math::
        \frac{1}{\sqrt{3}} \left[1, 1, 1 \right]^{\text{T}}.

    >>> from toqito.states import max_entangled
    >>> from toqito.state_ops import schmidt_decomposition
    >>> singular_vals, u_mat, vt_mat = schmidt_decomposition(max_entangled(3))
    >>> singular_vals
    [[0.57735027]
     [0.57735027]
     [0.57735027]]
    >>> u_mat
    [[1. 0. 0.]
     [0. 1. 0.]
     [0. 0. 1.]]
    >>> vt_mat
    [[1. 0. 0.]
     [0. 1. 0.]
     [0. 0. 1.]]

    References
    ==========
    .. [WikSD] Wikipedia: Schmidt decomposition
        https://en.wikipedia.org/wiki/Schmidt_decomposition

    :param vec: A bipartite quantum state to compute the Schmidt decomposition of.
    :param dim: An array consisting of the dimensions of the subsystems (default gives subsystems
                equal dimensions).
    :param k_param: How many terms of the Schmidt decomposition should be computed (default is 0).
    :return: The Schmidt decomposition of the :code:`vec` input.
    """
    eps = np.finfo(float).eps

    if dim is None:
        dim = np.round(np.sqrt(len(vec)))
    if isinstance(dim, list):
        dim = np.array(dim)

    # Allow the user to enter a single number for `dim`.
    if isinstance(dim, float):
        dim = np.array([dim, len(vec) / dim])
        if np.abs(dim[1] - np.round(dim[1])) >= 2 * len(vec) * eps:
            raise ValueError(
                "InvalidDim: The value of `dim` must evenly divide"
                " `len(vec)`; please provide a `dim` array "
                "containing the dimensions of the subsystems."
            )
        dim[1] = np.round(dim[1])

    # Try to guess whether SVD or SVDS will be faster, and then perform the
    # appropriate singular value decomposition.
    adj = 20 + 1000 * (not issparse(vec))

    # Just a few Schmidt coefficients.
    if 0 < k_param <= np.ceil(np.min(dim) / adj):
        u_mat, singular_vals, vt_mat = linalg.svds(
            linalg.LinearOperator(np.reshape(vec, dim[::-1].astype(int)), k_param)
        )
    # Otherwise, use lots of Schmidt coefficients.
    else:
        u_mat, singular_vals, vt_mat = np.linalg.svd(np.reshape(vec, dim[::-1].astype(int)))

    if k_param > 0:
        u_mat = u_mat[:, :k_param]
        singular_vals = singular_vals[:k_param]
        vt_mat = vt_mat[:, :k_param]

    # singular_vals = np.diag(singular_vals)
    singular_vals = singular_vals.reshape(-1, 1)
    if k_param == 0:
        # Schmidt rank.
        r_param = np.sum(singular_vals > np.max(dim) * np.spacing(singular_vals[0]))
        # Schmidt coefficients.
        singular_vals = singular_vals[:r_param]
        u_mat = u_mat[:, :r_param]
        vt_mat = vt_mat[:, :r_param]

    u_mat = u_mat.conj().T
    return singular_vals, u_mat, vt_mat



"""Generate random quantum states and measurements."""
from toqito.random.random_unitary import random_unitary
from toqito.random.random_density_matrix import random_density_matrix
from toqito.random.random_ginibre import random_ginibre
from toqito.random.random_povm import random_povm
from toqito.random.random_state_vector import random_state_vector



"""Generate Ginibre random matrix."""
import numpy as np


def random_ginibre(
    dim_n: int,
    dim_m: int,
) -> np.ndarray:
    r"""
    Generate a Ginibre random matrix [WIKCIRC]_.

    Generates a random :code:`dim_n`-by-:code:`dim_m` Ginibre matrix.

    A *Ginibre random matrix* is a matrix with independent and identically distributed complex
    standard Gaussian entries.

    Ginibre random matrices are used in the construction of Wishart-random POVMs [HMN20]_.

    Examples
    ==========

    Generate a random :math:`2`-by-:math:`2` Ginibre random matrix.

    >>> from toqito.random import random_ginibre
    >>> random_ginibre(2, 2)
    [[ 0.06037649-0.05158031j  0.46797859+0.21872729j]
     [-0.95223112-0.71959831j  0.3404352 +0.11166238j]]

    References
    ==========

    .. [WIKCIRC] Wikipedia: Circular law
        https://en.wikipedia.org/wiki/Circular_law

    .. [HMN20] Heinosaari, Teiko, Maria Anastasia Jivulescu, and Ion Nechita.
        "Random positive operator valued measures."
        Journal of Mathematical Physics 61.4 (2020): 042202.

    :param dim_n: The number of rows of the Ginibre random matrix.
    :param dim_m: The number of columns of the Ginibre random matrix.
    :return: A :code:`dim_n`-by-:code:`dim_m` Ginibre random density matrix.
    """
    return (np.random.randn(dim_n, dim_m) + 1j * np.random.randn(dim_n, dim_m)) / np.sqrt(2)



"""Perfect matchings."""
from typing import List, Union

import numpy as np


def perfect_matchings(num: Union[List[int], int, np.ndarray]) -> np.ndarray:
    r"""
    Give all perfect matchings of :code:`num` objects.

    The input can be either an even natural number (the number of objects to be matched) or a
    `numpy` array containing an even number of distinct objects to be matched.

    Returns all perfect matchings of a given list of objects. That is, it returns all ways of
    grouping an even number of objects into pairs.

    This function is adapted from QETLAB.

    Examples
    ==========

    This is an example of how to generate all perfect matchings of the numbers 1, 2, 3, 4.

    >>> from toqito.perms import perfect_matchings
    >>> perfect_matchings(4)
    [[1 2 3 4]
     [1 3 2 4]
     [1 4 3 2]]

    :param num: Either an even integer, indicating that you would like all perfect matchings of the
                integers 1,2, ... N, or a `list` or `np.array` containing an even number of distinct
                entries, indicating that you would like all perfect matchings of those entries.
    :return: An array containing all valid perfect matchings of size :code:`num`.
    """
    if isinstance(num, int):
        num = np.arange(1, num + 1)
    if isinstance(num, list):
        num = np.array(num)

    len_num = len(num)

    # Base case, `num = 2`: only one perfect matching.
    if len_num == 2:
        return num

    # There are no perfect matchings of an odd number of objects.
    if len_num % 2 == 1:
        return np.zeros((0, len_num))

    # Recursive step: build perfect matchings from smaller ones.

    # Only do the recursive step once instead of `num-1` times: we will then tweak
    # the output n-1 times.
    lower_fac = perfect_matchings(num[2:])
    if len(lower_fac.shape) == 1:
        lfac_size = 1
    else:
        lfac_size = lower_fac.shape[0]
    matchings = np.zeros((0, len_num), dtype=int)

    # Now build the perfect matchings we actually want.
    for j in range(1, len_num):
        tlower_fac = lower_fac.copy()
        tlower_fac[tlower_fac == num[j]] = num[1]

        one_vec = np.ones((lfac_size, 2), dtype=int) * [num[0], num[j]]
        if lfac_size == 1:
            one_vec = one_vec[0]

        s_vec = np.hstack((one_vec, tlower_fac))
        matchings = np.vstack((matchings, s_vec))

    return matchings



"""Is matrix a diagonal matrix."""
import numpy as np
from toqito.matrix_props import is_square


def is_diagonal(mat: np.ndarray) -> bool:
    r"""
    Determine if a matrix is diagonal [WikDiag]_.

    A matrix is diagonal if the matrix is square and if the diagonal of the matrix is non-zero,
    while the off-diagonal elements are all zero.

    The following is an example of a 3-by-3 diagonal matrix:

    .. math::
        \begin{equation}
            \begin{pmatrix}
                1 & 0 & 0 \\
                0 & 2 & 0 \\
                0 & 0 & 3
            \end{pmatrix}
        \end{equation}

    This quick implementation is given by Daniel F. from StackOverflow in [SODIA]_.

    Examples
    ==========

    Consider the following diagonal matrix:

    .. math::
        A = \begin{pmatrix}
                1 & 0 \\
                0 & 1
            \end{pmatrix}.

    Our function indicates that this is indeed a diagonal matrix:

    >>> from toqito.matrix_props import is_diagonal
    >>> import numpy as np
    >>> A = np.array([[1, 0], [0, 1]])
    >>> is_diagonal(A)
    True

    Alternatively, the following example matrix

    .. math::
        B = \begin{pmatrix}
                1 & 2 \\
                3 & 4
            \end{pmatrix}

    is not diagonal, as shown using :code:`toqito`.

    >>> from toqito.matrix_props import is_diagonal
    >>> import numpy as np
    >>> B = np.array([[1, 2], [3, 4]])
    >>> is_diagonal(B)
    False

    References
    ==========
    .. [WikDiag] Wikipedia: Diagonal matrix
        https://en.wikipedia.org/wiki/Diagonal_matrix

    .. [SODIA] StackOverflow post
        https://stackoverflow.com/questions/43884189/

    :param mat: The matrix to check.
    :return: Returns True if the matrix is diagonal and False otherwise.
    """
    if not is_square(mat):
        return False
    i, j = mat.shape
    test = mat.reshape(-1)[:-1].reshape(i - 1, j + 1)
    return ~np.any(test[:, 1:])



"""Two-player XOR game."""
import cvxpy
import numpy as np


class XORGame:
    r"""
    Create two-player XOR game object.

    Calculates the optimal probability that Alice and Bob win the game if they
    are allowed to determine a join strategy beforehand, but not allowed to
    communicate during the game itself.

    The quantum value of an XOR game can be solved via the semidefinite program
    from [CHTW04]_.

    This function is adapted from the QETLAB package.

    Examples
    ==========

    The CHSH game

    The CHSH game is a two-player nonlocal game with the following probability
    distribution and question and answer sets [CSUU08]_.

    .. math::
        \begin{equation}
            \begin{aligned} \pi(x,y) = \frac{1}{4}, \qquad (x,y) \in
                            \Sigma_A \times
                \Sigma_B, \qquad \text{and} \qquad (a, b) \in \Gamma_A \times
                \Gamma_B,
            \end{aligned}
        \end{equation}

    where

    .. math::
        \begin{equation}
            \Sigma_A = \{0, 1\}, \quad \Sigma_B = \{0, 1\}, \quad \Gamma_A =
            \{0,1\}, \quad \text{and} \quad \Gamma_B = \{0, 1\}.
        \end{equation}

    Alice and Bob win the CHSH game if and only if the following equation is
    satisfied

    .. math::
        \begin{equation}
        a \oplus b = x \land y.
        \end{equation}

    Recall that :math:`\oplus` refers to the XOR operation.

    The optimal quantum value of CHSH is :math:`\cos(\pi/8)^2 \approx 0.8536`
    where the optimal classical value is :math:`3/4`.

    In order to specify the CHSH game, we can define the probability matrix and
    predicate matrix for the CHSH game as `numpy` arrays as follows.

    >>> import numpy as np
    >>> prob_mat = np.array([[1 / 4, 1 / 4], [1 / 4, 1 / 4]])
    >>> pred_mat = np.array([[0, 0], [0, 1]])

    In `toqito`, we can calculate both the quantum and classical value of the
    CHSH game as follows.

    >>> import numpy as np
    >>> from toqito.nonlocal_games.xor_game import XORGame
    >>> chsh = XORGame(prob_mat, pred_mat)
    >>> chsh.quantum_value()
    0.8535533885683664
    >>>
    >>> chsh.classical_value()
    0.75

    The odd cycle game

    The odd cycle game is another XOR game [CHTW04]_. For this game, we can
    specify the probability and predicate matrices as follows.

    >>> prob_mat = np.array(
    >>> [
    >>>     [0.1, 0.1, 0, 0, 0],
    >>>     [0, 0.1, 0.1, 0, 0],
    >>>     [0, 0, 0.1, 0.1, 0],
    >>>     [0, 0, 0, 0.1, 0.1],
    >>>     [0.1, 0, 0, 0, 0.1],
    >>> ]
    >>> )
    >>> pred_mat = np.array(
    >>> [
    >>>     [0, 1, 0, 0, 0],
    >>>     [0, 0, 1, 0, 0],
    >>>     [0, 0, 0, 1, 0],
    >>>     [0, 0, 0, 0, 1],
    >>>     [1, 0, 0, 0, 0],
    >>> ]
    >>> )

    In :code:`toqito`, we can calculate both the quantum and classical value of
    the odd cycle game as follows.

    >>> import numpy as np
    >>> from toqito.nonlocal_games.xor_game import XORGame
    >>> odd_cycle = XORGame(prob_mat, pred_mat)
    >>> odd_cycle.quantum_value()
    0.9755282544736033
    >>> odd_cycle.classical_value()
    0.9

    References
    ==========
    .. [CSUU08] Richard Cleve, William Slofstra, Falk Unger, Sarvagya Upadhyay
        "Strong parallel repetition theorem for quantum XOR proof systems",
        2008,
        https://arxiv.org/abs/quant-ph/0608146

    .. [CHTW04] Richard Cleve, Peter Hoyer, Ben Toner, John Watrous
        "Consequences and limits of nonlocal strategies."
        Proceedings. 19th IEEE Annual Conference on Computational Complexity,
        IEEE, 2004.
        https://arxiv.org/abs/quant-ph/0404076
    """

    def __init__(
        self,
        prob_mat: np.ndarray,
        pred_mat: np.ndarray,
        reps: int = 1,
        tol: float = None,
    ) -> None:
        """
        Construct XOR game object.

        :param prob_mat: A matrix whose (q_0, q_1)-entry gives the probability that
                     the referee will give Alice the value `q_0` and Bob the
                     value `q_1`.
        :param pred_mat: A binary matrix whose (q_0, q_1)-entry indicates the
                     winning choice (either 0 or 1) when Alice and Bob receive
                     values `q_0` and `q_1` from the referee.
        :param tol: The error tolerance for the value.
        """
        self.prob_mat = prob_mat
        self.pred_mat = pred_mat
        self.reps = reps

        q_0, q_1 = self.prob_mat.shape
        if tol is None:
            self.tol = np.finfo(float).eps * q_0 ** 2 * q_1 ** 2
        else:
            self.tol = tol

        # Perform some basic error checking to ensure the probability and
        # predicate matrices are well-defined.
        if (q_0, q_1) != self.pred_mat.shape:
            raise ValueError(
                "Invalid: The matrices `prob_mat` and `pred_mat` must"
                " be matrices of the same size."
            )
        if -np.min(np.min(self.prob_mat)) > self.tol:
            raise ValueError(
                "Invalid: The variable `prob_mat` must be a "
                "probability matrix: its entries must be "
                "non-negative."
            )
        if np.abs(np.sum(np.sum(self.prob_mat)) - 1) > self.tol:
            raise ValueError(
                "Invalid: The variable `prob_mat` must be a "
                "probability matrix: its entries must sum to 1."
            )

    def quantum_value(self) -> float:
        r"""
        Compute the quantum value of the XOR game.

        To obtain the quantum value of the XOR game, we calculate the following
        simplified dual problem of the semidefinite program from the set of
        notes: https://cs.uwaterloo.ca/~watrous/CS867.Winter2017/Notes/06.pdf

        .. math::
            \begin{equation}
                \begin{aligned}
                    \text{minimize:} \quad & \frac{1}{2} \sum_{x \in X} u(x) +
                                             \frac{1}{2} \sum_{y \in Y} v(y) \\
                    \text{subject to:} \quad &
                            \begin{pmatrix}
                                \text{Diag}(u) & -D \\
                                -D^* & \text{Diag}(v)
                            \end{pmatrix} \geq 0, \\
                            & u \in \mathbb{R}^X, \
                              v \in \mathbb{R}^Y.
                \end{aligned}
            \end{equation}

        where :math:`D` is the matrix defined to be

        .. math::
            D(x,y) = \pi(x, y) (-1)^{f(x,y)}

        In other words, :math:`\pi(x, y)` corresponds to :code:`prob_mat[x, y]`,
        and :math:`f(x,y)` corresponds to :code:`pred_mat[x, y]`.

        :return: A value between [0, 1] representing the quantum value.
        """
        alice_in, bob_in = self.prob_mat.shape
        d_mat = np.zeros([alice_in, bob_in])

        for x_alice in range(alice_in):
            for y_bob in range(bob_in):
                d_mat[x_alice, y_bob] = self.prob_mat[x_alice, y_bob] * (-1) ** (
                    self.pred_mat[x_alice, y_bob]
                )

        u_vec = cvxpy.Variable(alice_in, complex=False)
        v_vec = cvxpy.Variable(bob_in, complex=False)

        objective = cvxpy.Minimize(cvxpy.sum(u_vec) + cvxpy.sum(v_vec))
        constraints = [
            cvxpy.bmat(
                [
                    [cvxpy.diag(u_vec), -d_mat],
                    [np.negative(d_mat.conj().T), cvxpy.diag(v_vec)],
                ]
            )
            >> 0
        ]

        problem = cvxpy.Problem(objective, constraints)
        problem.solve()

        if self.reps == 1:
            return np.real(problem.value) / 4 + 1 / 2
        # It holds from (https://arxiv.org/abs/quant-ph/0608146) that the
        # quantum value of any XOR game obeys strong parallel repetition. That
        # is, it holds that:
        #   \omega^*(G^{^n}) = \omega^*(G)^n,
        # where G^{^n} denotes playing the game G n-times.
        return (np.real(problem.value) / 4 + 1 / 2) ** self.reps

    def classical_value(self) -> float:
        """
        Compute the classical value of the XOR game.

        :return: A value between [0, 1] representing the classical value.
        """
        if self.reps == 1:
            q_0, q_1 = self.prob_mat.shape

            # At worst, out winning probability is 0. Now, try to improve.
            val = 0

            # Find the maximum probability of winning (this is NP-hard, so don't
            # expect an easy way to do it: just loop over all strategies.

            # Loop over Alice's answers
            for a_ans in range(2 ** q_0):
                # Loop over Bob's answers:
                for b_ans in range(2 ** q_1):
                    a_vec = (a_ans >> np.arange(q_0)) & 1
                    b_vec = (b_ans >> np.arange(q_1)) & 1

                    # Now compute the winning probability under this strategy:
                    # XOR together Alice's responses and Bob's responses, then
                    # check where the XORed value equals the value in the given
                    # matrix. Where the values match, multiply by the
                    # probability of getting that pair of questions (i.e.,
                    # multiply by the probability of getting that pair of
                    # questions (i.e., multiply entry-wise by P) and then sum
                    # over the rows and columns.
                    classical_strategy = np.mod(
                        np.multiply(a_vec.conj().T.reshape(-1, 1), np.ones((1, q_1)))
                        + np.multiply(np.ones((q_0, 1)), b_vec),
                        2,
                    )
                    p_win = np.sum(
                        np.sum(np.multiply(classical_strategy == self.pred_mat, self.prob_mat))
                    )
                    # Is this strategy better than other ones tried so far?
                    val = max(val, p_win)

                    # Already optimal? Quit.
                    if val >= 1 - self.tol:
                        return val
            return val
        raise ValueError(
            "Error: toqito currently does not support "
            "multiple repetitions for the classical value of "
            "a nonlocal game."
        )



"""Test gen_gell_mann."""
import numpy as np

from toqito.matrices import gen_gell_mann


def test_gell_mann_identity():
    """Generalized Gell-Mann operator identity."""
    expected_res = np.array([[1, 0], [0, 1]])
    res = gen_gell_mann(0, 0, 2)

    bool_mat = np.isclose(expected_res, res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_gell_mann_pauli_x():
    """Generalized Gell-Mann operator Pauli-X."""
    expected_res = np.array([[0, 1], [1, 0]])
    res = gen_gell_mann(0, 1, 2)

    bool_mat = np.isclose(expected_res, res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_gell_mann_pauli_y():
    """Generalized Gell-Mann operator Pauli-Y."""
    expected_res = np.array([[0, -1j], [1j, 0]])
    res = gen_gell_mann(1, 0, 2)

    bool_mat = np.isclose(expected_res, res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_gell_mann_pauli_z():
    """Generalized Gell-Mann operator Pauli-Z."""
    expected_res = np.array([[1, 0], [0, -1]])
    res = gen_gell_mann(1, 1, 2)

    bool_mat = np.isclose(expected_res, res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_gell_mann_dim_3_1():
    """Generalized Gell-Mann operator 3-dimensional."""
    expected_res = np.array([[0, 1, 0], [1, 0, 0], [0, 0, 0]])
    res = gen_gell_mann(0, 1, 3)

    bool_mat = np.isclose(expected_res, res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_gell_mann_dim_3_2():
    """Generalized Gell-Mann operator 3-dimensional."""
    expected_res = np.array([[0, 0, 1], [0, 0, 0], [1, 0, 0]])
    res = gen_gell_mann(0, 2, 3)

    bool_mat = np.isclose(expected_res, res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_gell_mann_dim_3_3():
    """Generalized Gell-Mann operator 3-dimensional."""
    expected_res = np.array(
        [
            [1 / np.sqrt(3), 0, 0],
            [0, 1 / np.sqrt(3), 0],
            [0, 0, -2 * 1 / np.sqrt(3)],
        ]
    )
    res = gen_gell_mann(2, 2, 3)

    bool_mat = np.isclose(expected_res, res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_gell_mann_dim_4_1():
    """Generalized Gell-Mann operator 4-dimensional."""
    expected_res = np.array([[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0]])
    res = gen_gell_mann(2, 3, 4)

    bool_mat = np.isclose(expected_res, res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_gell_mann_sparse_2():
    """Generalized Gell-Mann operator sparse."""
    res = gen_gell_mann(205, 34, 500, True)

    assert res[34, 205] == -1j
    assert res[205, 34] == 1j


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Test random_unitary."""
import numpy as np

from toqito.random import random_unitary
from toqito.matrix_props import is_unitary


def test_random_unitary_not_real():
    """Generate random non-real unitary matrix."""
    mat = random_unitary(2)
    np.testing.assert_equal(is_unitary(mat), True)


def test_random_unitary_real():
    """Generate random real unitary matrix."""
    mat = random_unitary(2, True)
    np.testing.assert_equal(is_unitary(mat), True)


def test_random_unitary_vec_dim():
    """Generate random non-real unitary matrix."""
    mat = random_unitary([4, 4], True)
    np.testing.assert_equal(is_unitary(mat), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Test ghz."""
import numpy as np

from toqito.states import basis
from toqito.states import ghz
from toqito.matrix_ops import tensor


def test_ghz_2_3():
    """Produces the 3-qubit GHZ state: `1/sqrt(2) * (|000> + |111>)`."""
    e_0, e_1 = basis(2, 0), basis(2, 1)
    expected_res = 1 / np.sqrt(2) * (tensor(e_0, e_0, e_0) + tensor(e_1, e_1, e_1))

    res = ghz(2, 3).toarray()

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_ghz_4_7():
    r"""
    The following generates the following GHZ state in `(C^4)^{\otimes 7}`.

    `1/sqrt(30) * (|0000000> + 2|1111111> + 3|2222222> + 4|3333333>)`.
    """
    e0_4 = np.array([[1], [0], [0], [0]])
    e1_4 = np.array([[0], [1], [0], [0]])
    e2_4 = np.array([[0], [0], [1], [0]])
    e3_4 = np.array([[0], [0], [0], [1]])

    expected_res = (
        1
        / np.sqrt(30)
        * (
            tensor(e0_4, e0_4, e0_4, e0_4, e0_4, e0_4, e0_4)
            + 2 * tensor(e1_4, e1_4, e1_4, e1_4, e1_4, e1_4, e1_4)
            + 3 * tensor(e2_4, e2_4, e2_4, e2_4, e2_4, e2_4, e2_4)
            + 4 * tensor(e3_4, e3_4, e3_4, e3_4, e3_4, e3_4, e3_4)
        )
    )

    res = ghz(4, 7, np.array([1, 2, 3, 4]) / np.sqrt(30)).toarray()

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_ghz_invalid_dim():
    """Tests for invalid dimensions."""
    with np.testing.assert_raises(ValueError):
        ghz(0, 2)


def test_ghz_invalid_qubits():
    """Tests for invalid number of qubits."""
    with np.testing.assert_raises(ValueError):
        ghz(2, 0)


def test_ghz_invalid_coeff():
    """Tests for invalid coefficients."""
    with np.testing.assert_raises(ValueError):
        ghz(2, 3, [1, 2, 3, 4, 5])


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Is matrix the identity matrix."""
import numpy as np
from toqito.matrix_props import is_square


def is_identity(mat: np.ndarray, rtol: float = 1e-05, atol: float = 1e-8) -> bool:
    r"""
    Check if matrix is the identity matrix [WikIdentity]_.

    For dimension :math:`n`, the :math:`n \times n` identity matrix is defined as

    .. math::
        I_n =
        \begin{pmatrix}
            1 & 0 & 0 & \ldots & 0 \\
            0 & 1 & 0 & \ldots & 0 \\
            0 & 0 & 1 & \ldots & 0 \\
            \vdots & \vdots & \vdots & \ddots & \vdots \\
            0 & 0 & 0 & \ldots & 1
        \end{pmatrix}.

    Examples
    ==========

    Consider the following matrix:

    .. math::
        A = \begin{pmatrix}
                1 & 0 & 0 \\
                0 & 1 & 0 \\
                0 & 0 & 1
            \end{pmatrix}

    our function indicates that this is indeed the identity matrix of dimension
    3.

    >>> from toqito.matrix_props import is_identity
    >>> import numpy as np
    >>> mat = np.eye(3)
    >>> is_identity(mat)
    True

    Alternatively, the following example matrix :math:`B` defined as

    .. math::
        B = \begin{pmatrix}
                1 & 2 & 3 \\
                4 & 5 & 6 \\
                7 & 8 & 9
            \end{pmatrix}

    is not an identity matrix.

    >>> from toqito.matrix_props import is_identity
    >>> import numpy as np
    >>> mat = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    >>> is_identity(mat)
    False

    References
    ==========
    .. [WikIdentity] Wikipedia: Identity matrix
        https://en.wikipedia.org/wiki/Identity_matrix

    :param mat: Matrix to check.
    :param rtol: The relative tolerance parameter (default 1e-05).
    :param atol: The absolute tolerance parameter (default 1e-08).
    :return: Return :code:`True` if matrix is the identity matrix, and
            :code:`False` otherwise.
    """
    if not is_square(mat):
        return False
    id_mat = np.eye(len(mat))
    return np.allclose(mat, id_mat, rtol=rtol, atol=atol)



"""Properties of quantum states."""
from toqito.state_props.is_ensemble import is_ensemble
from toqito.state_props.is_pure import is_pure
from toqito.state_props.is_mixed import is_mixed
from toqito.state_props.is_mutually_orthogonal import is_mutually_orthogonal
from toqito.state_props.is_mutually_unbiased_basis import is_mutually_unbiased_basis
from toqito.state_props.is_ppt import is_ppt
from toqito.state_props.is_product_vector import is_product_vector
from toqito.state_props.concurrence import concurrence
from toqito.state_props.negativity import negativity
from toqito.state_props.log_negativity import log_negativity
from toqito.state_props.purity import purity
from toqito.state_props.schmidt_rank import schmidt_rank
from toqito.state_props.von_neumann_entropy import von_neumann_entropy
from toqito.state_props.entanglement_of_formation import entanglement_of_formation
from toqito.state_props.l1_norm_coherence import l1_norm_coherence
from toqito.state_props.in_separable_ball import in_separable_ball
from toqito.state_props.has_symmetric_extension import has_symmetric_extension



"""Generalized Pauli matrices."""
import numpy as np

from toqito.matrices import shift
from toqito.matrices import clock


def gen_pauli(k_1: int, k_2: int, dim: int) -> np.ndarray:
    r"""
    Produce generalized Pauli operator [WikGenPaul]_.

    Generates a `dim`-by-`dim` unitary operator. More specifically, it is the
    operator `X^k_1*Z^k_2`, where X and Z are the "shift" and "clock" operators
    that naturally generalize the Pauli X and Z operators. These matrices span
    the entire space of `dim`-by-`dim` matrices as `k_1` and `k_2` range from 0
    to `dim-1`, inclusive.

    Note that the generalized Pauli operators are also known by the name of
    "discrete Weyl operators". [WatrousLec6]_

    Examples
    ==========

    The generalized Pauli operator for `k_1 = 1`, `k_2 = 0` and `dim = 2` is
    given as the standard Pauli-X matrix

    .. math::
        G_{1, 0, 2} = \begin{pmatrix}
                         0 & 1 \\
                         1 & 0
                      \end{pmatrix}.

    This can be obtained in `toqito` as follows.

    >>> from toqito.matrices import gen_pauli
    >>> dim = 2
    >>> k_1 = 1
    >>> k_2 = 0
    >>> gen_pauli(k_1, k_2, dim)
    [[0.+0.j, 1.+0.j],
     [1.+0.j, 0.+0.j]])

    The generalized Pauli matrix `k_1 = 1`, `k_2 = 1`, and `dim = 2` is given as
    the standard Pauli-Y matrix

    .. math::
        G_{1, 1, 2} = \begin{pmatrix}
                        0 & -1 \\
                        1 & 0
                      \end{pmatrix}.

    This can be obtained in `toqito` as follows.`

    >>> from toqito.matrices import gen_pauli
    >>> dim = 2
    >>> k_1 = 1
    >>> k_2 = 1
    >>> gen_pauli(k_1, k_2, dim)
    [[ 0.+0.0000000e+00j, -1.+1.2246468e-16j],
     [ 1.+0.0000000e+00j,  0.+0.0000000e+00j]])

    References
    ==========
    .. [WikGenPaul] Wikipedia: Generalizations of Pauli matrices
        https://en.wikipedia.org/wiki/Generalizations_of_Pauli_matrices

    .. [WatrousLec6] Lecture 6: Further remarks on measurements and channels
        https://cs.uwaterloo.ca/~watrous/LectureNotes/CS766.Fall2011/06.pdf

    :param k_1: (a non-negative integer from 0 to `dim-1` inclusive).
    :param k_2: (a non-negative integer from 0 to `dim-1` inclusive).
    :param dim: (a positive integer indicating the dimension).
    :return: A generalized Pauli operator.
    """
    gen_pauli_x = shift(dim)
    gen_pauli_z = clock(dim)

    gen_pauli_w = np.linalg.matrix_power(gen_pauli_x, k_1) @ np.linalg.matrix_power(
        gen_pauli_z, k_2
    )

    return gen_pauli_w



"""State exclusion."""
from typing import List

import cvxpy
import numpy as np

from .state_helper import __is_states_valid, __is_probs_valid


def state_exclusion(
    states: List[np.ndarray], probs: List[float] = None, method: str = "conclusive"
) -> float:
    r"""
    Compute probability of single state exclusion.

    The *quantum state exclusion* problem involves a collection of :math:`n`
    quantum states

    .. math::
        \rho = \{ \rho_0, \ldots, \rho_n \},

    as well as a list of corresponding probabilities

    .. math::
        p = \{ p_0, \ldots, p_n \}.

    Alice chooses :math:`i` with probability :math:`p_i` and creates the state
    :math:`\rho_i`.

    Bob wants to guess which state he was *not* given from the collection of
    states. State exclusion implies that ability to discard (with certainty) at
    least one out of the "n" possible quantum states by applying a measurement.

    This function implements the following semidefinite program that provides
    the optimal probability with which Bob can conduct quantum state exclusion.

        .. math::
            \begin{equation}
                \begin{aligned}
                    \text{minimize:} \quad & \sum_{i=0}^n p_i \langle M_i,
                                                \rho_i \rangle \\
                    \text{subject to:} \quad & M_0 + \ldots + M_n =
                                               \mathbb{I}, \\
                                             & M_0, \ldots, M_n >= 0.
                \end{aligned}
            \end{equation}

    The conclusive state exclusion SDP is written explicitly in [BJOP14]_. The problem of conclusive
    state exclusion was also thought about under a different guise in [PBR12]_.


    For the unambiguous case, the following semidefinite program that provides
    the optimal probability with which Bob can conduct quantum state exclusion.

    .. math::
        \begin{align*}
            \text{maximize:} \quad & \sum_{i=0}^n \sum_{j=0}^n
                                     \langle M_i, \rho_j \rangle \\
            \text{subject to:} \quad & \sum_{i=0}^n M_i \leq \mathbb{I},\\
                                     & \text{Tr}(\rho_i M_i) = 0,
                                       \quad \quad \forall 1  \leq i \leq n, \\
                                     & M_0, \ldots, M_n \geq 0
        \end{align*}

    Examples
    ==========

    Consider the following two Bell states

    .. math::
        \begin{equation}
            \begin{aligned}
                u_0 &= \frac{1}{\sqrt{2}} \left( |00 \rangle + |11 \rangle \right), \\
                u_1 &= \frac{1}{\sqrt{2}} \left( |00 \rangle - |11 \rangle \right).
            \end{aligned}
        \end{equation}

    For the corresponding density matrices :math:`\rho_0 = u_0 u_0^*` and
    :math:`\rho_1 = u_1 u_1^*`, we may construct a set

    .. math::
        \rho = \{\rho_0, \rho_1 \}

    such that

    .. math::
        p = \{1/2, 1/2\}.

    It is not possible to conclusively exclude either of the two states. We can see that the result
    of the function in :code:`toqito` yields a value of :math:`0` as the probability for this to
    occur.

    >>> from toqito.state_opt import state_exclusion
    >>> from toqito.states import bell
    >>> import numpy as np
    >>> rho1 = bell(0) * bell(0).conj().T
    >>> rho2 = bell(1) * bell(1).conj().T
    >>>
    >>> states = [rho1, rho2]
    >>> probs = [1/2, 1/2]
    >>>
    >>> state_exclusion(states, probs, "conclusive")
    1.6824720366950206e-09

    Consider the following two Bell states

    .. math::
        \begin{equation}
            \begin{aligned}
                u_0 &= \frac{1}{\sqrt{2}} \left( |00 \rangle + |11 \rangle \right) \\
                u_1 &= \frac{1}{\sqrt{2}} \left( |00 \rangle - |11 \rangle \right).
            \end{aligned}
        \end{equation}

    For the corresponding density matrices :math:`\rho_0 = u_0 u_0^*` and
    :math:`\rho_1 = u_1 u_1^*`, we may construct a set

    .. math::
        \rho = \{\rho_0, \rho_1 \}

    such that

    .. math::
        p = \{1/2, 1/2\}.

    It is not possible to unambiguously exclude either of the two states. We can see that the result
    of the function in :code:`toqito` yields a value of :math:`0` as the probability for this to
    occur.

    >>> from toqito.state_opt import state_exclusion
    >>> from toqito.states import bell
    >>> import numpy as np
    >>> rho1 = bell(0) * bell(0).conj().T
    >>> rho2 = bell(1) * bell(1).conj().T
    >>>
    >>> states = [rho1, rho2]
    >>> probs = [1/2, 1/2]
    >>>
    >>> state_exclusion(states, probs, "unambiguous")
    -7.250173600116328e-18

    References
    ==========
    .. [PBR12] "On the reality of the quantum state"
        Pusey, Matthew F., Barrett, Jonathan, and Rudolph, Terry.
        Nature Physics 8.6 (2012): 475-478.
        arXiv:1111.3328

    .. [BJOP14] "Conclusive exclusion of quantum states"
        Bandyopadhyay, Somshubhro, Jain, Rahul, Oppenheim, Jonathan,
        Perry, Christopher
        Physical Review A 89.2 (2014): 022336.
        arXiv:1306.4683

    :param states: A list of states provided as either matrices or vectors.
    :param probs: Respective list of probabilities each state is selected.
    :param method: Exclusion method (either `conclusive` or `unambiguous`.
    :return: The optimal probability with which Bob can guess the state he was
             not given from `states`.
    """
    obj_func = []
    measurements = []
    constraints = []

    __is_states_valid(states)
    if probs is None:
        probs = [1 / len(states)] * len(states)
    __is_probs_valid(probs)

    supported_methods = ["conclusive", "unambiguous"]
    if method not in supported_methods:
        raise ValueError(
            f"Exclusion method {method} not supported. Please "
            f"select one from {supported_methods}."
        )

    dim_x, dim_y = states[0].shape

    # The variable `states` is provided as a list of vectors. Transform them
    # into density matrices.
    if dim_y == 1:
        for i, state_ket in enumerate(states):
            states[i] = state_ket * state_ket.conj().T

    for i, _ in enumerate(states):
        measurements.append(cvxpy.Variable((dim_x, dim_x), PSD=True))

        obj_func.append(probs[i] * cvxpy.trace(states[i].conj().T @ measurements[i]))

        if method == "unambiguous":
            constraints.append(cvxpy.trace(states[i] @ measurements[i]) == 0)

    if method == "conclusive":
        constraints.append(sum(measurements) == np.identity(dim_x))
    elif method == "unambiguous":
        constraints.append(sum(measurements) <= np.identity(dim_x))

    if method == "conclusive":
        if np.iscomplexobj(states[0]):
            objective = cvxpy.Minimize(cvxpy.real(sum(obj_func)))
        else:
            objective = cvxpy.Minimize(sum(obj_func))
    elif method == "unambiguous":
        if np.iscomplexobj(states[0]):
            objective = cvxpy.Maximize(cvxpy.real(sum(obj_func)))
        else:
            objective = cvxpy.Maximize(sum(obj_func))

    problem = cvxpy.Problem(objective, constraints)
    sol_default = problem.solve()

    return 1 / len(states) * sol_default



"""Is matrix a normal matrix."""
import numpy as np

from toqito.matrix_props import is_square


def is_normal(mat: np.ndarray, rtol: float = 1e-05, atol: float = 1e-08) -> bool:
    r"""
    Determine if a matrix is normal [WikNormal]_.

    A matrix is normal if it commutes with its adjoint

    .. math::
        \begin{equation}
            [X, X^*] = 0,
        \end{equation}

    or, equivalently if

    .. math::
        \begin{equation}
            X^* X = X X^*
        \end{equation}.

    Examples
    ==========

    Consider the following matrix

    .. math::
        A = \begin{pmatrix}
                1 & 0 & 0 & 0 \\
                0 & 1 & 0 & 0 \\
                0 & 0 & 1 & 0 \\
                0 & 0 & 0 & 1
            \end{pmatrix}

    our function indicates that this is indeed a normal matrix.

    >>> from toqito.matrix_props import is_normal
    >>> import numpy as np
    >>> A = np.identity(4)
    >>> is_normal(A)
    True

    Alternatively, the following example matrix :math:`B` defined as

    .. math::
        B = \begin{pmatrix}
                1 & 2 & 3 \\
                4 & 5 & 6 \\
                7 & 8 & 9
            \end{pmatrix}

    is not normal.

    >>> from toqito.matrix_props import is_normal
    >>> import numpy as np
    >>> B = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    >>> is_normal(B)
    False

    References
    ==========
    .. [WikNormal] Wikipedia: Normal matrix.
        https://en.wikipedia.org/wiki/Normal_matrix

    :param mat: The matrix to check.
    :param rtol: The relative tolerance parameter (default 1e-05).
    :param atol: The absolute tolerance parameter (default 1e-08).
    :return: Returns :code:`True` if the matrix is normal and :code:`False` otherwise.
    """
    if not is_square(mat):
        return False
    return np.allclose(mat @ mat.conj().T, mat.conj().T @ mat, rtol=rtol, atol=atol)



"""Gisin state."""
import numpy as np


def gisin(lambda_var: float, theta: float) -> np.ndarray:
    r"""
    Produce a Gisin state [GIS96]_.

    Returns the Gisin state described in [GIS96]_. Specifically, the Gisin state can be defined as:

    .. math::
        \begin{equation}
            \rho_{\lambda, \theta} = \lambda
                                    \begin{pmatrix}
                                        0 & 0 & 0 & 0 \\
                                        0 & \sin^2(\theta) &
                                        -\sin(\theta)\cos(\theta) & 0 \\
                                        0 & -\sin(\theta)\cos(\theta) &
                                        \cos^2(\theta) & 0 \\
                                        0 & 0 & 0 & 0
                                    \end{pmatrix} +
                                    \frac{1 - \lambda}{2}
                                    \begin{pmatrix}
                                        1 & 0 & 0 & 0 \\
                                        0 & 0 & 0 & 0 \\
                                        0 & 0 & 0 & 0 \\
                                        0 & 0 & 0 & 1
                                    \end{pmatrix}.
        \end{equation}

    Examples
    ==========

    The following code generates the Gisin state :math:`\rho_{0.5, 1}`.

    >>> from toqito.states import gisin
    >>> gisin(0.5, 1)
    [[ 0.25      ,  0.        ,  0.        ,  0.        ],
     [ 0.        ,  0.35403671, -0.22732436,  0.        ],
     [ 0.        , -0.22732436,  0.14596329,  0.        ],
     [ 0.        ,  0.        ,  0.        ,  0.25      ]]

    References
    ==========
    .. [GIS96] N. Gisin.
        Hidden quantum nonlocality revealed by local filters.
        (http://dx.doi.org/10.1016/S0375-9601(96)80001-6). 1996.

    :param lambda_var: A real parameter in [0, 1].
    :param theta: A real parameter.
    :return: Gisin state.
    """
    if lambda_var < 0 or lambda_var > 1:
        raise ValueError("InvalidLambda: Variable lambda must be between 0 and 1.")

    rho_theta = np.array(
        [
            [0, 0, 0, 0],
            [0, np.sin(theta) ** 2, -np.sin(2 * theta) / 2, 0],
            [0, -np.sin(2 * theta) / 2, np.cos(theta) ** 2, 0],
            [0, 0, 0, 0],
        ]
    )

    rho_uu_dd = np.array([[1, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 1]])

    return lambda_var * rho_theta + (1 - lambda_var) * rho_uu_dd / 2



"""Test state_exclusion."""
import numpy as np

from toqito.state_opt import state_exclusion
from toqito.states import bell


def test_conclusive_state_exclusion_one_state():
    """Conclusive state exclusion for single state."""
    rho = bell(0) * bell(0).conj().T
    states = [rho]

    res = state_exclusion(states, probs=None, method="conclusive")
    np.testing.assert_equal(np.isclose(res, 1), True)


def test_conclusive_state_exclusion_one_state_vec():
    """Conclusive state exclusion for single vector state."""
    rho = bell(0)
    states = [rho]

    res = state_exclusion(states, probs=None, method="conclusive")
    np.testing.assert_equal(np.isclose(res, 1), True)


def test_conclusive_state_exclusion_three_state():
    """Conclusive state exclusion for three Bell state density matrices."""
    rho1 = bell(0) * bell(0).conj().T
    rho2 = bell(1) * bell(1).conj().T
    rho3 = bell(2) * bell(2).conj().T
    states = [rho1, rho2, rho3]
    probs = [1 / 3, 1 / 3, 1 / 3]

    res = state_exclusion(states, probs, "conclusive")
    np.testing.assert_equal(np.isclose(res, 0), True)


def test_conclusive_state_exclusion_three_state_vec():
    """Conclusive state exclusion for three Bell state vectors."""
    rho1 = bell(0)
    rho2 = bell(1)
    rho3 = bell(2)
    states = [rho1, rho2, rho3]
    probs = [1 / 3, 1 / 3, 1 / 3]

    res = state_exclusion(states, probs, "conclusive")
    np.testing.assert_equal(np.isclose(res, 0), True)


def test_conclusive_state_exclusion_complex_three_state_vec():
    """Conclusive state exclusion on complex set of pure states."""
    mat_1 = np.array(
        [
            [0.65925619 + 0.0j, 0.38049979 + 0.01670518j, 0.26366168 - 0.1003037j],
            [0.38049979 - 0.01670518j, 0.22003457 + 0.0j, 0.14963473 - 0.06457285j],
            [0.26366168 + 0.1003037j, 0.14963473 + 0.06457285j, 0.12070924 + 0.0j],
        ]
    )

    mat_2 = np.array(
        [
            [0.36978742 + 0.0j, 0.2010124 - 0.07736818j, 0.42433574 - 0.08119137j],
            [0.2010124 + 0.07736818j, 0.12545538 + 0.0j, 0.24765141 + 0.04464623j],
            [0.42433574 + 0.08119137j, 0.24765141 - 0.04464623j, 0.5047572 + 0.0j],
        ]
    )

    mat_3 = np.array(
        [
            [0.34592805 + 0.0j, 0.28877002 + 0.03437698j, 0.31590575 + 0.20468388j],
            [0.28877002 - 0.03437698j, 0.24447252 + 0.0j, 0.28404902 + 0.13947028j],
            [0.31590575 - 0.20468388j, 0.28404902 - 0.13947028j, 0.40959944 + 0.0j],
        ]
    )
    states = [mat_1, mat_2, mat_3]
    probs = [1 / 3, 1 / 3, 1 / 3]

    res = state_exclusion(states, probs, "conclusive")
    assert res > 0


def test_unambiguous_state_exclusion_one_state():
    """Unambiguous state exclusion for single state."""
    mat = bell(0) * bell(0).conj().T
    states = [mat]

    res = state_exclusion(states, probs=None, method="unambiguous")
    np.testing.assert_equal(np.isclose(res, 0), True)


def test_unambiguous_state_exclusion_one_state_vec():
    """Unambiguous state exclusion for single vector state."""
    vec = bell(0)
    states = [vec]

    res = state_exclusion(states, probs=None, method="unambiguous")
    np.testing.assert_equal(np.isclose(res, 0), True)


def test_unambiguous_state_exclusion_three_state():
    """Unambiguous state exclusion for three Bell state density matrices."""
    mat1 = bell(0) * bell(0).conj().T
    mat2 = bell(1) * bell(1).conj().T
    mat3 = bell(2) * bell(2).conj().T
    states = [mat1, mat2, mat3]
    probs = [1 / 3, 1 / 3, 1 / 3]

    res = state_exclusion(states, probs, method="unambiguous")
    np.testing.assert_equal(np.isclose(res, 0), True)


def test_unambiguous_state_exclusion_three_state_vec():
    """Unambiguous state exclusion for three Bell state vectors."""
    mat1 = bell(0)
    mat2 = bell(1)
    mat3 = bell(2)
    states = [mat1, mat2, mat3]
    probs = [1 / 3, 1 / 3, 1 / 3]

    res = state_exclusion(states, probs, method="unambiguous")
    np.testing.assert_equal(np.isclose(res, 0), True)


def test_unambiguous_state_exclusion_complex_three_state_vec():
    """Unambiguous state exclusion on complex set of pure states."""
    mat_1 = np.array(
        [
            [0.37166502 + 0.0j, 0.06990262 + 0.00381928j, 0.44548935 - 0.17369055j],
            [0.06990262 - 0.00381928j, 0.01318651 + 0.0j, 0.0820026 - 0.03724556j],
            [0.44548935 + 0.17369055j, 0.0820026 + 0.03724556j, 0.61514848 + 0.0j],
        ]
    )

    mat_2 = np.array(
        [
            [0.03351844 + 0.0j, 0.08195346 + 0.02701392j, 0.15185457 + 0.0434629j],
            [0.08195346 - 0.02701392j, 0.22215 + 0.0j, 0.40631695 - 0.01611805j],
            [0.15185457 - 0.0434629j, 0.40631695 + 0.01611805j, 0.74433156 + 0.0j],
        ]
    )

    mat_3 = np.array(
        [
            [0.51449115 + 0.0j, 0.23369567 + 0.15751255j, 0.41173315 - 0.02901644j],
            [0.23369567 - 0.15751255j, 0.15437363 + 0.0j, 0.17813679 - 0.13923301j],
            [0.41173315 + 0.02901644j, 0.17813679 + 0.13923301j, 0.33113522 + 0.0j],
        ]
    )
    states = [mat_1, mat_2, mat_3]
    probs = [1 / 3, 1 / 3, 1 / 3]

    res = state_exclusion(states, probs, method="unambiguous")
    np.testing.assert_equal(np.isclose(res, 0, atol=1e-06), True)


def test_state_exclusion_invalid_method():
    """Invalid exclusion method for state exclusion."""
    with np.testing.assert_raises(ValueError):
        mat1 = bell(0)
        mat2 = bell(1)
        mat3 = bell(2)
        states = [mat1, mat2, mat3]
        probs = [1 / 3, 1 / 3, 1 / 3]

        state_exclusion(states, probs, method="not_valid")


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Swap."""
from typing import List, Union

import numpy as np

from toqito.perms import permute_systems


def swap(
    rho: np.ndarray,
    sys: List[int] = None,
    dim: Union[List[int], List[List[int]], int, np.ndarray] = None,
    row_only: bool = False,
) -> np.ndarray:
    r"""
    Swap two subsystems within a state or operator.

    Swaps the two subsystems of the vector or matrix :code:`rho`, where the dimensions of the
    (possibly more than 2) subsystems are given by :code:`dim` and the indices of the two subsystems
    to be swapped are specified in the 1-by-2 vector :code:`sys`.

    If :code:`rho` is non-square and not a vector, different row and column dimensions can be
    specified by putting the row dimensions in the first row of :code:`dim` and the column
    dimensions in the second row of :code:`dim`.

    If :code:`row_only` is set to :code:`True`, then only the rows of :code:`rho` are swapped, but
    not the columns -- this is equivalent to multiplying :code:`rho` on the left by the
    corresponding swap operator, but not on the right.

    Examples
    ==========

    Consider the following matrix

    .. math::
        X =
        \begin{pmatrix}
            1 & 5 & 9 & 13 \\
            2 & 6 & 10 & 14 \\
            3 & 7 & 11 & 15 \\
            4 & 8 & 12 & 16
        \end{pmatrix}.

    If we apply the :code:`swap` function provided by :code:`toqito` on :math:`X`, we should obtain
    the following matrix

    .. math::
        \text{Swap}(X) =
        \begin{pmatrix}
            1 & 9 & 5 & 13 \\
            3 & 11 & 7 & 15 \\
            2 & 10 & 6 & 14 \\
            4 & 12 & 8 & 16
        \end{pmatrix}.

    This can be observed by the following example in :code:`toqito`.

    >>> from toqito.perms import swap
    >>> import numpy as np
    >>> test_mat = np.array(
    >>>     [[1, 5, 9, 13], [2, 6, 10, 14], [3, 7, 11, 15], [4, 8, 12, 16]]
    >>> )
    >>> swap(test_mat)
    [[ 1  9  5 13]
     [ 3 11  7 15]
     [ 2 10  6 14]
     [ 4 12  8 16]]

    It is also possible to use the :code:`sys` and :code:`dim` arguments, it is possible to specify
    the system and dimension on which to apply the swap operator. For instance for
    :code:`sys = [1 ,2]` and :code:`dim = 2` we have that

    .. math::
        Swap(X)_{2, [1, 2]} =
        \begin{pmatrix}
            1 & 9 & 5 & 13 \\
            3 & 11 & 7 & 15 \\
            2 & 10 & 6 & 14 \\
            4 & 12 & 8 & 16
        \end{pmatrix}.

    Using :code:`toqito` we can see this gives the proper result.

    >>> from toqito.perms import swap
    >>> import numpy as np
    >>> test_mat = np.array(
    >>>     [[1, 5, 9, 13], [2, 6, 10, 14], [3, 7, 11, 15], [4, 8, 12, 16]]
    >>> )
    >>> swap(test_mat, [1, 2], 2)
    [[ 1  9  5 13]
     [ 3 11  7 15]
     [ 2 10  6 14]
     [ 4 12  8 16]]

    It is also possible to perform the :code:`swap` function on vectors in addition to matrices.

    >>> from toqito.perms import swap
    >>> import numpy as np
    >>> test_vec = np.array([1, 2, 3, 4])
    >>> swap(test_vec)
    [1 3 2 4]

    :param rho: A vector or matrix to have its subsystems swapped.
    :param sys: Default: [1, 2]
    :param dim: Default: :code:`[sqrt(len(X), sqrt(len(X)))]`
    :param row_only: Default: :code:False
    :return: The swapped matrix.
    """
    eps = np.finfo(float).eps
    if len(rho.shape) == 1:
        rho_dims = (1, rho.shape[0])
    else:
        rho_dims = rho.shape

    round_dim = np.round(np.sqrt(rho_dims))

    if sys is None:
        sys = [1, 2]

    if isinstance(dim, list):
        dim = np.array(dim)
    if dim is None:
        dim = np.array([[round_dim[0], round_dim[0]], [round_dim[1], round_dim[1]]])

    if isinstance(dim, int):
        dim = np.array([[dim, rho_dims[0] / dim], [dim, rho_dims[1] / dim]])
        if (
            np.abs(dim[0, 1] - np.round(dim[0, 1])) + np.abs(dim[1, 1] - np.round(dim[1, 1]))
            >= 2 * np.prod(rho_dims) * eps
        ):
            val_error = """
                InvalidDim: The value of `dim` must evenly divide the number of
                rows and columns of `rho`; please provide the `dim` array
                containing the dimensions of the subsystems.
            """
            raise ValueError(val_error)

        dim[0, 1] = np.round(dim[0, 1])
        dim[1, 1] = np.round(dim[1, 1])
        num_sys = 2
    else:
        num_sys = len(dim)

    # Verify that the input sys makes sense.
    if any(sys) < 1 or any(sys) > num_sys:
        val_error = """
            InvalidSys: The subsystems in `sys` must be between 1 and 
            `len(dim).` inclusive.
        """
        raise ValueError(val_error)
    if len(sys) != 2:
        val_error = """
            InvalidSys: `sys` must be a vector with exactly two elements.
        """
        raise ValueError(val_error)

    # Swap the indicated subsystems.
    perm = list(range(1, num_sys + 1))
    perm[sys[0] - 1 :] = perm[sys[0] - 1 :][::-1]
    return permute_systems(rho, perm, dim, row_only)



"""Clock matrix."""
from cmath import exp, pi

import numpy as np


def clock(dim: int) -> np.ndarray:
    r"""
    Produce clock matrix [WikClock]_.

    Returns the clock matrix of dimension `dim` described in [WikClock]_. The
    clock matrix generates the following `dim`-by-`dim` matrix

    .. math::
        \Sigma_{1, d} = \begin{pmatrix}
                        1 & 0 & 0 & \ldots & 0 \\
                        0 & \omega & 0 & \ldots & 0 \\
                        0 & 0 & \omega^2 & \ldots & 0 \\
                        \vdots & \vdots & \vdots & \ddots & \vdots \\
                        0 & 0 & 0 & \ldots & \omega^{d-1}
                   \end{pmatrix}

    where :math:`\omega` is the n-th primitive root of unity.

    The clock matrix is primarily used in the construction of the generalized
    Pauli operators.

    Examples
    ==========

    The clock matrix generated from :math:`d = 3` yields the following matrix:

    .. math::
        \Sigma_{1, 3} = \begin{pmatrix}
            1 & 0 & 0 \\
            0 & \omega & 0 \\
            0 & 0 & \omega^2
        \end{pmatrix}

    >>> from toqito.matrices import clock
    >>> clock(3)
    [[ 1. +0.j       ,  0. +0.j       ,  0. +0.j       ],
     [ 0. +0.j       , -0.5+0.8660254j,  0. +0.j       ],
     [ 0. +0.j       ,  0. +0.j       , -0.5-0.8660254j]]

    References
    ==========
    .. [WikClock] Wikipedia: Generalizations of Pauli matrices,
        https://en.wikipedia.org/wiki/Generalizations_of_Pauli_matrices

    :param dim: Dimension of the matrix.
    :return: `dim`-by-`dim` clock matrix.
    """
    c_var = 2j * pi / dim
    omega = (exp(k * c_var) for k in range(dim))
    return np.diag(list(omega))



"""Test breuer."""
import numpy as np

from toqito.states import breuer


def test_breuer_dim_4_ent():
    """Generate Breuer state of dimension 4 with weight 0.1."""
    expected_res = np.array([[0.3, 0, 0, 0], [0, 0.2, 0.1, 0], [0, 0.1, 0.2, 0], [0, 0, 0, 0.3]])

    res = breuer(2, 0.1)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_breuer_invalid_dim():
    """Ensures that an odd dimension if not accepted."""
    with np.testing.assert_raises(ValueError):
        breuer(3, 0.1)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Is matrix a positive semidefinite matrix."""
import numpy as np

from toqito.matrix_props import is_hermitian


def is_positive_semidefinite(mat: np.ndarray, rtol: float = 1e-05, atol: float = 1e-08) -> bool:
    r"""
    Check if matrix is positive semidefinite (PSD) [WikPSD]_.

    Examples
    ==========

    Consider the following matrix

    .. math::
        A = \begin{pmatrix}
                1 & -1 \\
                -1 & 1
            \end{pmatrix}

    our function indicates that this is indeed a positive semidefinite matrix.

    >>> from toqito.matrix_props import is_positive_semidefinite
    >>> import numpy as np
    >>> A = np.array([[1, -1], [-1, 1]])
    >>> is_positive_semidefinite(A)
    True

    Alternatively, the following example matrix :math:`B` defined as

    .. math::
        B = \begin{pmatrix}
                -1 & -1 \\
                -1 & -1
            \end{pmatrix}

    is not positive semidefinite.

    >>> from toqito.matrix_props import is_positive_semidefinite
    >>> import numpy as np
    >>> B = np.array([[-1, -1], [-1, -1]])
    >>> is_positive_semidefinite(B)
    False

    References
    ==========
    .. [WikPSD] Wikipedia: Definiteness of a matrix.
        https://en.wikipedia.org/wiki/Definiteness_of_a_matrix

    :param mat: Matrix to check.
    :param rtol: The relative tolerance parameter (default 1e-05).
    :param atol: The absolute tolerance parameter (default 1e-08).
    :return: Return :code:`True` if matrix is PSD, and :code:`False` otherwise.
    """
    if not is_hermitian(mat, rtol, atol):
        return False
    evals, _ = np.linalg.eigh(mat)
    return all(x >= -abs(atol) for x in evals)



"""Calculates success probability of approximately cloning a quantum state."""
from typing import List, Union
import cvxpy
import numpy as np

from toqito.matrix_ops import tensor
from toqito.channels import partial_trace
from toqito.perms import permutation_operator


def optimal_clone(
    states: List[np.ndarray],
    probs: List[float],
    num_reps: int = 1,
    strategy: bool = False,
) -> Union[float, np.ndarray]:
    r"""
    Compute probability of counterfeiting quantum money [MVW12]_.

    The primal problem for the :math:`n`-fold parallel repetition is given as follows:

    .. math::
        \begin{equation}
            \begin{aligned}
                \text{maximize:} \quad &
                \langle W_{\pi} \left(Q^{\otimes n} \right) W_{\pi}^*, X \rangle \\
                \text{subject to:} \quad & \text{Tr}_{\mathcal{Y}^{\otimes n}
                                           \otimes \mathcal{Z}^{\otimes n}}(X)
                                           = \mathbb{I}_{\mathcal{X}^{\otimes
                                           n}},\\
                                           & X \in \text{Pos}(
                                           \mathcal{Y}^{\otimes n}
                                           \otimes \mathcal{Z}^{\otimes n}
                                           \otimes \mathcal{X}^{\otimes n}).
            \end{aligned}
        \end{equation}

    The dual problem for the :math:`n`-fold parallel repetition is given as follows:

    .. math::
        \begin{equation}
            \begin{aligned}
                \text{minimize:} \quad & \text{Tr}(Y) \\
                \text{subject to:} \quad & \mathbb{I}_{\mathcal{Y}^{\otimes n}
                \otimes \mathcal{Z}^{\otimes n}} \otimes Y \geq W_{\pi}
                \left( Q^{\otimes n} \right) W_{\pi}^*, \\
                & Y \in \text{Herm} \left(\mathcal{X}^{\otimes n} \right).
            \end{aligned}
        \end{equation}

    Examples
    ==========

    Wiesner's original quantum money scheme [Wies83]_ was shown in [MVW12]_ to have an optimal
    probability of 3/4 for succeeding a counterfeiting attack.

    Specifically, in the single-qubit case, Wiesner's quantum money scheme corresponds to the
    following ensemble:

    .. math::
        \left\{
            \left( \frac{1}{4}, |0\rangle \right),
            \left( \frac{1}{4}, |1\rangle \right),
            \left( \frac{1}{4}, |+\rangle \right),
            \left( \frac{1}{4}, |-\rangle \right)
        \right\},

    which yields the operator

    .. math::
        \begin{equation}
            Q = \frac{1}{4} \left(|000 \rangle \langle 000| + |111 \rangle \langle 111| +
                                  |+++ \rangle + \langle +++| + |--- \rangle \langle ---| \right).
        \end{equation}

    We can see that the optimal value we obtain in solving the SDP is 3/4.

    >>> from toqito.state_opt import optimal_clone
    >>> from toqito.states import basis
    >>> import numpy as np
    >>> e_0, e_1 = basis(2, 0), basis(2, 1)
    >>> e_p = (e_0 + e_1) / np.sqrt(2)
    >>> e_m = (e_0 - e_1) / np.sqrt(2)
    >>>
    >>> states = [e_0, e_1, e_p, e_m]
    >>> probs = [1 / 4, 1 / 4, 1 / 4, 1 / 4]
    >>> wiesner = optimal_clone(states, probs)
    0.749999999967631

    References
    ==========
    .. [MVW12] Abel Molina, Thomas Vidick, and John Watrous.
        "Optimal counterfeiting attacks and generalizations for Wiesner’s
        quantum money."
        Conference on Quantum Computation, Communication, and Cryptography.
        Springer, Berlin, Heidelberg, 2012.
        https://arxiv.org/abs/1202.4010

    .. [Wies83] Stephen Wiesner
        "Conjugate coding."
        ACM Sigact News 15.1 (1983): 78-88.
        https://dl.acm.org/doi/pdf/10.1145/1008908.1008920

    :param states: A list of states provided as either matrices or vectors.
    :param probs: Respective list of probabilities each state is selected.
    :param num_reps: Number of parallel repetitions to perform.
    :param strategy: Boolean that denotes whether to return strategy.
    :return: The optimal probability with of counterfeiting quantum money.
    """
    dim = len(states[0]) ** 3

    # Construct the following operator:
    #                                ___               ___
    # Q = ∑_{k=1}^N p_k |ψ_k ⊗ ψ_k ⊗ ψ_k> <ψ_k ⊗ ψ_k ⊗ ψ_k|
    q_a = np.zeros((dim, dim))
    for k, state in enumerate(states):
        q_a += (
            probs[k]
            * tensor(state, state, state.conj())
            * tensor(state, state, state.conj()).conj().T
        )

    # The system is over:
    # Y_1 ⊗ Z_1 ⊗ X_1, ... , Y_n ⊗ Z_n ⊗ X_n.
    num_spaces = 3

    # In the event of more than a single repetition, one needs to apply a
    # permutation operator to the variables in the SDP to properly align
    # the spaces.
    if num_reps == 1:
        pperm = np.array([1])
    else:
        # The permutation vector `perm` contains elements of the
        # sequence from: https://oeis.org/A023123
        q_a = tensor(q_a, num_reps)
        perm = []
        for i in range(1, num_spaces + 1):
            perm.append(i)
            var = i
            for j in range(1, num_reps):
                perm.append(var + num_spaces * j)
        pperm = permutation_operator(2, perm)

    if strategy:
        return primal_problem(q_a, pperm, num_reps)
    return dual_problem(q_a, pperm, num_reps)


def primal_problem(q_a: np.ndarray, pperm: np.ndarray, num_reps: int) -> float:
    """
    Primal problem for counterfeit attack.

    As the primal problem takes longer to solve than the dual problem (as
    the variables are of larger dimension), the primal problem is only here
    for reference.

    :return: The optimal value of performing a counterfeit attack.
    """
    num_spaces = 3

    sys = list(range(1, num_spaces * num_reps))
    sys = [elem for elem in sys if elem % num_spaces != 0]

    # The dimension of each subsystem is assumed to be of dimension 2.
    dim = 2 * np.ones((1, num_spaces * num_reps)).astype(int).flatten()
    dim = dim.tolist()

    x_var = cvxpy.Variable((8 ** num_reps, 8 ** num_reps), hermitian=True)
    if num_reps == 1:
        objective = cvxpy.Maximize(cvxpy.trace(cvxpy.real(q_a.conj().T @ x_var)))
    else:
        objective = cvxpy.Maximize(
            cvxpy.trace(cvxpy.real(pperm @ q_a.conj().T @ pperm.conj().T @ x_var))
        )
    constraints = [
        partial_trace(x_var, sys, dim) == np.identity(2 ** num_reps),
        x_var >> 0,
    ]
    problem = cvxpy.Problem(objective, constraints)

    return problem.solve()


def dual_problem(q_a: np.ndarray, pperm: np.ndarray, num_reps: int) -> float:
    """
    Dual problem for counterfeit attack.

    :return: The optimal value of performing a counterfeit attack.
    """
    y_var = cvxpy.Variable((2 ** num_reps, 2 ** num_reps), hermitian=True)
    objective = cvxpy.Minimize(cvxpy.trace(cvxpy.real(y_var)))

    kron_var = cvxpy.kron(cvxpy.kron(np.eye(2 ** num_reps), np.eye(2 ** num_reps)), y_var)

    if num_reps == 1:
        constraints = [cvxpy.real(kron_var) >> q_a]
    else:
        constraints = [cvxpy.real(kron_var) >> pperm @ q_a @ pperm.conj().T]
    problem = cvxpy.Problem(objective, constraints)

    return problem.solve()



"""Tests for purity."""
import numpy as np

from toqito.states import werner
from toqito.state_props import purity


def test_purity():
    """Test for identity matrix."""
    expected_res = 1 / 4
    res = purity(np.identity(4) / 4)
    np.testing.assert_equal(res, expected_res)


def test_purity_non_density_matrix():
    r"""Test purity on non-density matrix."""
    rho = np.array([[1, 2], [3, 4]])

    with np.testing.assert_raises(ValueError):
        purity(rho)


def test_purity_werner_state():
    """Test purity of mixed Werner state."""
    res = purity(werner(2, 1 / 4))
    np.testing.assert_equal(np.isclose(res, 0.2653, atol=4), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Is matrix commuting."""
import numpy as np


def is_commuting(mat_1: np.ndarray, mat_2: np.ndarray) -> bool:
    r"""
    Determine if two linear operators commute with each other [WikCom]_.

    For any pair of operators :math:`X, Y \in \text{L}(\mathcal{X})`, the
    Lie bracket :math:`\left[X, Y\right] \in \text{L}(\mathcal{X})` is defined
    as

    .. math::
        \left[X, Y\right] = XY - YX.

    It holds that :math:`` if and only if :math:`X` and :math:`Y` commute
    [WatCom18]_.

    Examples
    ==========

    Consider the following matrices:

    .. math::
        A = \begin{pmatrix} 0 & 1 \\ 0 & 0 \end{pmatrix},
        \quad \text{and} \quad
        B = \begin{pmatrix} 1 & 0 \\ 0 & 0 \end{pmatrix}.

    It holds that :math:`AB=0`, however

    .. math::
        BA = \begin{pmatrix} 0 & 1 \\ 0 & 0 \end{pmatrix} = A,

    and hence, do not commute.

    >>> from toqito.matrix_props import is_commuting
    >>> import numpy as np
    >>> mat_1 = np.array([[0, 1], [0, 0]])
    >>> mat_2 = np.array([[1, 0], [0, 0]])
    >>> is_commuting(mat_1, mat_2)
    False

    Consider the following pair of matrices:

    .. math::
        A = \begin{pmatrix}
            1 & 0 & 0 \\
            0 & 1 & 0 \\
            1 & 0 & 2
            \end{pmatrix}, \quad \text{and} \quad
        B = \begin{pmatrix}
            2 & 4 & 0 \\
            3 & 1 & 0 \\
            -1 & -4 & 1
            \end{pmatrix}.

    It may be verified that :math:`AB = BA = 0`, and therefore :math`A` and
    :math:`B` commute.

    >>> from toqito.matrix_props import is_commuting
    >>> import numpy as np
    >>> mat_1 = np.array([[1, 0, 0], [0, 1, 0], [1, 0, 2]])
    >>> mat_2 = np.array([[2, 4, 0], [3, 1, 0], [-1, -4, 1]])
    >>> is_commuting(mat_1, mat_2)
    True

    References
    ==========
    .. [WikCom] Wikipedia: Commuting matrices
        https://en.wikipedia.org/wiki/Commuting_matrices

    .. [WatCom18] Watrous, John.
        "The theory of quantum information."
        Section: "Lie brackets and commutants".
        Cambridge University Press, 2018.

    :param mat_1: First matrix to check.
    :param mat_2: Second matrix to check.
    :return: Return `True` if :code:`mat_1` commutes with :code:`mat_2` and False otherwise.
    """
    return np.allclose(mat_1 @ mat_2 - mat_2 @ mat_1, 0)



"""Permute systems."""
import functools
import operator

from typing import List, Union
from scipy import sparse

import numpy as np

from toqito.matrix_ops import vec


def permute_systems(
    input_mat: np.ndarray,
    perm: Union[np.ndarray, List[int]],
    dim: Union[np.ndarray, List[int]] = None,
    row_only: bool = False,
    inv_perm: bool = False,
) -> np.ndarray:
    r"""
    Permute subsystems within a state or operator.

    Permutes the order of the subsystems of the vector or matrix :code:`input_mat` according to the
    permutation vector :code:`perm`, where the dimensions of the subsystems are given by the vector
    :code:`dim`. If :code:`input_mat` is non-square and not a vector, different row and column
    dimensions can be specified by putting the row dimensions in the first row of :code:`dim` and
    the columns dimensions in the second row of :code:`dim`.

    If :code:`row_only = True`, then only the rows of :code:`input_mat` are permuted, but not the
    columns -- this is equivalent to multiplying :code:`input_mat` on the left by the corresponding
    permutation operator, but not on the right.

    If :code:`row_only = False`, then :code:`dim` only needs to contain the row dimensions of the
    subsystems, even if :code:`input_mat` is not square. If :code:`inv_perm = True`, then the
    inverse permutation of :code:`perm` is applied instead of :code:`perm` itself.

    Examples
    ==========

    For spaces :math:`\mathcal{A}` and :math:`\mathcal{B}` where
    :math:`\text{dim}(\mathcal{A}) = \text{dim}(\mathcal{B}) = 2` we may consider an operator
    :math:`X \in \mathcal{A} \otimes \mathcal{B}`. Applying the `permute_systems` function with
    vector :math:`[2,1]` on :math:`X`, we may reorient the spaces such that
    :math:`X \in \mathcal{B} \otimes \mathcal{A}`.

    For example, if we define :math:`X \in \mathcal{A} \otimes \mathcal{B}` as

    .. math::
        X = \begin{pmatrix}
            1 & 2 & 3 & 4 \\
            5 & 6 & 7 & 8 \\
            9 & 10 & 11 & 12 \\
            13 & 14 & 15 & 16
        \end{pmatrix},

    then applying the `permute_systems` function on :math:`X` to obtain
    :math:`X \in \mathcal{B} \otimes \mathcal{A}` yield the following matrix

    .. math::
        X_{[2,1]} = \begin{pmatrix}
            1 & 3 & 2 & 4 \\
            9 & 11 & 10 & 12 \\
            5 & 7 & 6 & 8 \\
            13 & 15 & 14 & 16
        \end{pmatrix}.

    >>> from toqito.perms import permute_systems
    >>> import numpy as np
    >>> test_input_mat = np.array(
    >>>    [[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]]
    >>> )
    >>> permute_systems(test_input_mat, [2, 1])
    [[ 1  3  2  4]
     [ 9 11 10 12]
     [ 5  7  6  8]
     [13 15 14 16]]

    For spaces :math:`\mathcal{A}, \mathcal{B}`, and :math:`\mathcal{C}` where
    :math:`\text{dim}(\mathcal{A}) = \text{dim}(\mathcal{B}) = \text{dim}(\mathcal{C}) = 2` we may
    consider an operator :math:`X \in \mathcal{A} \otimes \mathcal{B} \otimes \mathcal{C}`. Applying
    the :code:`permute_systems` function with vector :math:`[2,3,1]` on :math:`X`, we may reorient
    the spaces such that :math:`X \in \mathcal{B} \otimes \mathcal{C} \otimes \mathcal{A}`.

    For example, if we define :math:`X \in \mathcal{A} \otimes \mathcal{B} \otimes \mathcal{C}` as

    .. math::
        X =
        \begin{pmatrix}
            1 & 2 & 3 & 4, 5 & 6 & 7 & 8 \\
            9 & 10 & 11 & 12 & 13 & 14 & 15 & 16 \\
            17 & 18 & 19 & 20 & 21 & 22 & 23 & 24 \\
            25 & 26 & 27 & 28 & 29 & 30 & 31 & 32 \\
            33 & 34 & 35 & 36 & 37 & 38 & 39 & 40 \\
            41 & 42 & 43 & 44 & 45 & 46 & 47 & 48 \\
            49 & 50 & 51 & 52 & 53 & 54 & 55 & 56 \\
            57 & 58 & 59 & 60 & 61 & 62 & 63 & 64
        \end{pmatrix},

    then applying the `permute_systems` function on :math:`X` to obtain
    :math:`X \in \mathcal{B} \otimes \mathcal{C} \otimes \mathcal{C}` yield the following matrix

    .. math::
        X_{[2, 3, 1]} =
        \begin{pmatrix}
            1 & 5 & 2 & 6 & 3 & 7 & 4, 8 \\
            33 & 37 & 34 & 38 & 35 & 39 & 36 & 40 \\
            9 & 13 & 10 & 14 & 11 & 15 & 12 & 16 \\
            41 & 45 & 42 & 46 & 43 & 47 & 44 & 48 \\
            17 & 21 & 18 & 22 & 19 & 23 & 20 & 24 \\
            49 & 53 & 50 & 54 & 51 & 55 & 52 & 56 \\
            25 & 29 & 26 & 30 & 27 & 31 & 28 & 32 \\
            57 & 61 & 58 & 62 & 59 & 63 & 60 & 64
        \end{pmatrix}.

    >>> from toqito.perms import permute_systems
    >>> import numpy as np
    >>> test_input_mat = np.array(
    >>>    [
    >>>        [1, 2, 3, 4, 5, 6, 7, 8],
    >>>        [9, 10, 11, 12, 13, 14, 15, 16],
    >>>        [17, 18, 19, 20, 21, 22, 23, 24],
    >>>        [25, 26, 27, 28, 29, 30, 31, 32],
    >>>        [33, 34, 35, 36, 37, 38, 39, 40],
    >>>        [41, 42, 43, 44, 45, 46, 47, 48],
    >>>        [49, 50, 51, 52, 53, 54, 55, 56],
    >>>        [57, 58, 59, 60, 61, 62, 63, 64],
    >>>    ]
    >>> )
    >>> permute_systems(test_input_mat, [2, 3, 1])
    [[ 1  5  2  6  3  7  4  8]
     [33 37 34 38 35 39 36 40]
     [ 9 13 10 14 11 15 12 16]
     [41 45 42 46 43 47 44 48]
     [17 21 18 22 19 23 20 24]
     [49 53 50 54 51 55 52 56]
     [25 29 26 30 27 31 28 32]
     [57 61 58 62 59 63 60 64]]

    :param input_mat: The vector or matrix.
    :param perm: A permutation vector.
    :param dim: The default has all subsystems of equal dimension.
    :param row_only: Default: :code:`False`
    :param inv_perm: Default: :code:`True`
    :return: The matrix or vector that has been permuted.
    """
    if len(input_mat.shape) == 1:
        input_mat_dims = (1, input_mat.shape[0])
    else:
        input_mat_dims = input_mat.shape

    is_vec = np.min(input_mat_dims) == 1
    num_sys = len(perm)

    if dim is None:
        x_tmp = input_mat_dims[0] ** (1 / num_sys) * np.ones(num_sys)
        y_tmp = input_mat_dims[1] ** (1 / num_sys) * np.ones(num_sys)
        dim = np.array([x_tmp, y_tmp])

    if isinstance(dim, list):
        dim = np.array(dim)

    if is_vec:
        # 1 if column vector
        if len(input_mat.shape) > 1:
            vec_orien = 0
        # 2 if row vector
        elif len(input_mat.shape) == 1:
            vec_orien = 1
        else:
            raise ValueError(
                "InvalidMat: Length of tuple of dimensions "
                "specifying the input matrix can only be of "
                "length 1 or length 2."
            )

    if len(dim.shape) == 1:
        # Force dim to be a row vector.
        dim_tmp = dim[:].T
        if is_vec:
            dim = np.ones((2, len(dim)))
            dim[vec_orien, :] = dim_tmp
        else:
            dim = np.array([[dim_tmp], [dim_tmp]])

    prod_dim_r = int(np.prod(dim[0, :]))
    prod_dim_c = int(np.prod(dim[1, :]))

    if len(perm) != num_sys:
        raise ValueError("InvalidPerm: `len(perm)` must be equal to " "`len(dim)`.")
    if sorted(perm) != list(range(1, num_sys + 1)):
        raise ValueError("InvalidPerm: `perm` must be a permutation vector.")
    if input_mat_dims[0] != prod_dim_r or (not row_only and input_mat_dims[1] != prod_dim_c):
        raise ValueError(
            "InvalidDim: The dimensions specified in DIM do not " "agree with the size of X."
        )
    if is_vec:
        if inv_perm:
            permuted_mat_1 = input_mat.reshape(dim[vec_orien, ::-1].astype(int), order="F")
            permuted_mat = vec(np.transpose(permuted_mat_1, num_sys - np.array(perm[::-1]))).T
            # We need to flatten out the array.
            permuted_mat = functools.reduce(operator.iconcat, permuted_mat, [])
        else:
            permuted_mat_1 = input_mat.reshape(dim[vec_orien, ::-1].astype(int), order="F")
            permuted_mat = vec(np.transpose(permuted_mat_1, num_sys - np.array(perm[::-1]))).T
            # We need to flatten out the array.
            permuted_mat = functools.reduce(operator.iconcat, permuted_mat, [])
        return np.array(permuted_mat)

    vec_arg = np.array(list(range(0, input_mat_dims[0])))

    # If the dimensions are specified, ensure they are given to the
    # recursive calls as flattened lists.
    if len(dim[0][:]) == 1:
        dim = functools.reduce(operator.iconcat, dim, [])

    row_perm = permute_systems(vec_arg, perm, dim[0][:], False, inv_perm)

    # This condition is only necessary if the `input_mat` variable is sparse.
    if isinstance(input_mat, (sparse.csr_matrix, sparse.dia_matrix)):
        input_mat = input_mat.toarray()
        permuted_mat = input_mat[row_perm, :]
        permuted_mat = np.array(permuted_mat)
    else:
        permuted_mat = input_mat[row_perm, :]

    if not row_only:
        vec_arg = np.array(list(range(0, input_mat_dims[1])))
        col_perm = permute_systems(vec_arg, perm, dim[1][:], False, inv_perm)
        permuted_mat = permuted_mat[:, col_perm]

    return permuted_mat



"""Test is_unitary."""
import numpy as np

from toqito.matrix_props import is_unitary
from toqito.random import random_unitary


def test_is_unitary_random():
    """Test that random unitary matrix returns True."""
    mat = random_unitary(2)
    np.testing.assert_equal(is_unitary(mat), True)


def test_is_unitary_hardcoded():
    """Test that hardcoded unitary matrix returns True."""
    mat = np.array([[0, 1], [1, 0]])
    np.testing.assert_equal(is_unitary(mat), True)


def test_is_not_unitary():
    """Test that non-unitary matrix returns False."""
    mat = np.array([[1, 0], [1, 1]])
    np.testing.assert_equal(is_unitary(mat), False)


def test_is_not_unitary_matrix():
    """Test that non-unitary matrix returns False."""
    mat = np.array([[1, 0], [1, 1]])
    np.testing.assert_equal(is_unitary(mat), False)


def test_is_unitary_not_square():
    """Input must be a square matrix."""
    mat = np.array([[-1, 1, 1], [1, 2, 3]])
    np.testing.assert_equal(is_unitary(mat), False)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Generate random POVM."""
import numpy as np


def random_povm(dim: int, num_inputs: int, num_outputs: int) -> np.ndarray:
    """
    Generate random positive operator valued measurements (POVMs) [WIKPOVM]_.

    Examples
    ==========

    We can generate a set of POVMs consisting of a specific dimension along with a given number of
    measurement inputs and measurement outputs. As an example, we can construct a random set of
    POVMs of dimension :math:`2` with :math:`2` inputs and :math:`2` outputs.

    >>> from toqito.random import random_povm
    >>> import numpy as np
    >>>
    >>> dim, num_inputs, num_outputs = 2, 2, 2
    >>> povms = random_povm(dim, num_inputs, num_outputs)
    >>> povms
    [[[[ 0.40313832+0.j,  0.59686168+0.j],
       [ 0.91134633+0.j,  0.08865367+0.j]],
     [[-0.27285707+0.j,  0.27285707+0.j],
      [-0.12086852+0.j,  0.12086852+0.j]]],
     [[[-0.27285707+0.j,  0.27285707+0.j],
      [-0.12086852+0.j,  0.12086852+0.j]],
     [[ 0.452533  +0.j,  0.547467  +0.j],
      [ 0.34692158+0.j,  0.65307842+0.j]]]]

    We can verify that this constitutes a valid set of POVM elements as checking that these
    operators all sum to the identity operator.

    >>> np.round(povms[:, :, 0, 0] + povms[:, :, 0, 1])
    [[1.+0.j, 0.+0.j],
     [0.+0.j, 1.+0.j]]

    References
    ==========
    .. [WIKPOVM] Wikipedia: POVM
        https://en.wikipedia.org/wiki/POVM

    :param dim: The dimension of the measurements.
    :param num_inputs: The number of inputs for the measurement.
    :param num_outputs: The number of outputs for the measurement.
    :return: A set of POVMs of dimension :code:`dim`.
    """
    povms = []
    gram_vectors = np.random.normal(size=(dim, dim, num_inputs, num_outputs))
    for input_block in gram_vectors:
        normalizer = sum(
            [np.array(output_block).T.conj() @ output_block for output_block in input_block]
        )

        u_mat, d_mat, _ = np.linalg.svd(normalizer)

        output_povms = []
        for output_block in input_block:
            partial = (
                np.array(output_block, dtype=complex).dot(u_mat).dot(np.diag(d_mat ** (-1 / 2.0)))
            )
            internal = partial.dot(np.diag(np.ones(dim)) ** (1 / 2.0))
            output_povms.append(internal.T.conj() @ internal)
        povms.append(output_povms)

    # This allows us to index the POVMs as [d, d, num_inputs, num_outputs].
    povms = np.swapaxes(np.array(povms), 0, 2)
    povms = np.swapaxes(povms, 1, 3)

    return povms



"""Check if state is pure."""
from typing import List, Union

import numpy as np


def is_pure(state: Union[List[np.ndarray], np.ndarray]) -> bool:
    r"""
    Determine if a given state is pure or list of states are pure [WikIsPure]_.

    A state is said to be pure if it is a density matrix with rank equal to 1. Equivalently, the
    state :math:`\rho` is pure if there exists a unit vector :math:`u` such that:

    .. math::
        \rho = u u^*.

    Examples
    ==========

    Consider the following Bell state:

    .. math::
        u = \frac{1}{\sqrt{2}} \left( |00 \rangle + |11 \rangle \right) \in \mathcal{X}.

    The corresponding density matrix of :math:`u` may be calculated by:

    .. math::
        \rho = u u^* = \frac{1}{2} \begin{pmatrix}
                         1 & 0 & 0 & 1 \\
                         0 & 0 & 0 & 0 \\
                         0 & 0 & 0 & 0 \\
                         1 & 0 & 0 & 1
                       \end{pmatrix} \in \text{D}(\mathcal{X}).

    Calculating the rank of :math:`\rho` yields that the :math:`\rho` is a pure state. This can be
    confirmed in :code:`toqito` as follows:

    >>> from toqito.states import bell
    >>> from toqito.state_props import is_pure
    >>> u = bell(0)
    >>> rho = u * u.conj().T
    >>> is_pure(rho)
    True

    It is also possible to determine whether a set of density matrices are pure. For instance, we
    can see that the density matrices corresponding to the four Bell states yield a result of
    :code:`True` indicating that all states provided to the function are pure.

    >>> from toqito.states import bell
    >>> from toqito.state_props import is_pure
    >>> u0, u1, u2, u3 = bell(0), bell(1), bell(2), bell(3)
    >>> rho0 = u0 * u0.conj().T
    >>> rho1 = u1 * u1.conj().T
    >>> rho2 = u2 * u2.conj().T
    >>> rho3 = u3 * u3.conj().T
    >>>
    >>> is_pure([rho0, rho1, rho2, rho3])
    True

    References
    ==========
    .. [WikIsPure] Wikipedia: Quantum state - Pure states
        https://en.wikipedia.org/wiki/Quantum_state#Pure_states

    :param state: The density matrix representing the quantum state or a list
                  of density matrices representing quantum states.
    :return: :code:`True` if state is pure and :code:`False` otherwise.
    """
    # Allow the user to enter a list of states to check.
    if isinstance(state, list):
        for rho in state:
            eigs, _ = np.linalg.eig(rho)
            if not np.allclose(np.max(np.diag(eigs)), 1):
                return False
        return True

    # Otherwise, if the user just put in a single state, check that.
    eigs, _ = np.linalg.eig(state)
    return np.allclose(np.max(np.diag(eigs)), 1)



"""Negativity property."""
from typing import List, Union

import numpy as np

from toqito.state_ops import pure_to_mixed
from toqito.channels import partial_transpose


def negativity(rho: np.ndarray, dim: Union[List[int], int] = None) -> float:
    r"""
    Compute the negativity of a bipartite quantum state [WikNeg]_.

    The negativity of a subsystem can be defined in terms of a density matrix :math:`\rho`:

    .. math::
        \mathcal{N}(\rho) \equiv \frac{||\rho^{\Gamma_A}||_1-1}{2}.

    Calculate the negativity of the quantum state :math:`\rho`, assuming that the two subsystems on
    which :math:`\rho` acts are of equal dimension (if the local dimensions are unequal, specify
    them in the optional :code:`dim` argument). The negativity of :math:`\rho` is the sum of the
    absolute value of the negative eigenvalues of the partial transpose of :math:`\rho`.

    Examples
    ==========

    Example of the negativity of density matrix of Bell state.

    >>> from toqito.states import bell
    >>> from toqito.state_props import negativity
    >>> rho = bell(0) * bell(0).conj().T
    >>> negativity(rho)
    0.4999999999999998

    See Also
    ==========
    log_negativity

    References
    ==========
    .. [WikNeg] Wikipedia page for negativity (quantum mechanics):
        https://en.wikipedia.org/wiki/Negativity_(quantum_mechanics)

    :param rho: A density matrix of a pure state vector.
    :param dim: The default has both subsystems of equal dimension.
    :return: A value between 0 and 1 that corresponds to the negativity of :math:`\rho`.
    """
    # Allow the user to input either a pure state vector or a density matrix.
    rho = pure_to_mixed(rho)
    rho_dims = rho.shape
    round_dim = np.round(np.sqrt(rho_dims))

    if dim is None:
        dim = np.array([round_dim])
        dim = dim.T
    if isinstance(dim, list):
        dim = np.array(dim)

    # Allow the user to enter a single number for dim.
    if isinstance(dim, int):
        dim = np.array([dim, rho_dims[0] / dim])
        if abs(dim[1] - np.round(dim[1])) >= 2 * rho_dims[0] * np.finfo(float).eps:
            raise ValueError(
                "InvalidDim: If `dim` is a scalar, `rho` must be "
                "square and `dim` must evenly divide `len(rho)`. "
                "Please provide the `dim` array containing the "
                "dimensions of the subsystems."
            )
        dim[1] = np.round(dim[1])

    if np.prod(dim) != rho_dims[0]:
        raise ValueError(
            "InvalidDim: Please provide local dimensions in the "
            "argument `dim` that match the size of `rho`."
        )

    # Compute the negativity.
    return (np.linalg.norm(partial_transpose(rho, 2, dim), ord="nuc") - 1) / 2



"""Test is_pure."""
import numpy as np

from toqito.state_props import is_pure
from toqito.states import basis
from toqito.states import bell


def test_is_pure_state():
    """Ensure that pure Bell state returns True."""
    rho = bell(0) * bell(0).conj().T
    np.testing.assert_equal(is_pure(rho), True)


def test_is_pure_list():
    """Check that list of pure states returns True."""
    e_0, e_1, e_2 = basis(3, 0), basis(3, 1), basis(3, 2)

    e0_dm = e_0 * e_0.conj().T
    e1_dm = e_1 * e_1.conj().T
    e2_dm = e_2 * e_2.conj().T

    np.testing.assert_equal(is_pure([e0_dm, e1_dm, e2_dm]), True)


def test_is_pure_not_pure_state():
    """Check that non-pure state returns False."""
    rho = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    np.testing.assert_equal(is_pure(rho), False)


def test_is_pure_not_pure_list():
    """Check that list of non-pure states return False."""
    rho = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    sigma = np.array([[1, 2, 3], [10, 11, 12], [7, 8, 9]])
    np.testing.assert_equal(is_pure([rho, sigma]), False)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Test partial_transpose."""
import cvxpy
import numpy as np

from cvxpy.atoms.affine.vstack import Vstack

from toqito.channels import partial_transpose
from toqito.states import bell


def test_partial_transpose():
    """
    Default partial_transpose.

    By default, the partial_transpose function performs the transposition
    on the second subsystem.
    """
    test_input_mat = np.arange(1, 17).reshape(4, 4)

    expected_res = np.array([[1, 5, 3, 7], [2, 6, 4, 8], [9, 13, 11, 15], [10, 14, 12, 16]])

    res = partial_transpose(test_input_mat)

    bool_mat = np.isclose(expected_res, res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_partial_transpose_sys():
    """
    Default partial transpose `sys` argument.

    By specifying the `sys` argument, you can perform the transposition on
    the first subsystem instead:
    """
    test_input_mat = np.arange(1, 17).reshape(4, 4)

    expected_res = np.array([[1, 2, 9, 10], [5, 6, 13, 14], [3, 4, 11, 12], [7, 8, 15, 16]])

    res = partial_transpose(test_input_mat, 1)

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_partial_transpose_sys_vec():
    """Partial transpose on matrix with `sys` defined as vector."""
    test_input_mat = np.arange(1, 17).reshape(4, 4)

    expected_res = np.array([[1, 5, 9, 13], [2, 6, 10, 14], [3, 7, 11, 15], [4, 8, 12, 16]])

    res = partial_transpose(test_input_mat, [1, 2])

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_partial_transpose_sys_vec_dim_vec():
    """Variables `sys` and `dim` defined as vector."""
    test_input_mat = np.arange(1, 17).reshape(4, 4)

    expected_res = np.array([[1, 5, 9, 13], [2, 6, 10, 14], [3, 7, 11, 15], [4, 8, 12, 16]])

    res = partial_transpose(test_input_mat, [1, 2], [2, 2])

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_partial_transpose_norm_diff():
    """
    Apply partial transpose to first and second subsystem.

    Applying the transpose to both the first and second subsystems results
    in the standard transpose of the matrix.
    """
    test_input_mat = np.arange(1, 17).reshape(4, 4)
    res = np.linalg.norm(partial_transpose(test_input_mat, [1, 2]) - test_input_mat.conj().T)
    expected_res = 0

    np.testing.assert_equal(np.isclose(res, expected_res), True)


def test_partial_transpose_16_by_16():
    """Partial transpose on a 16-by-16 matrix."""
    test_input_mat = np.arange(1, 257).reshape(16, 16)
    res = partial_transpose(test_input_mat, [1, 3], [2, 2, 2, 2])
    first_expected_row = np.array(
        [1, 2, 33, 34, 5, 6, 37, 38, 129, 130, 161, 162, 133, 134, 165, 166]
    )

    first_expected_col = np.array([1, 17, 3, 19, 65, 81, 67, 83, 9, 25, 11, 27, 73, 89, 75, 91])

    np.testing.assert_equal(np.allclose(res[0, :], first_expected_row), True)
    np.testing.assert_equal(np.allclose(res[:, 0], first_expected_col), True)


def test_partial_transpose_bell_state():
    """Test partial transpose on a Bell state."""
    rho = bell(2) * bell(2).conj().T
    expected_res = np.array(
        [[0, 0, 0, 1 / 2], [0, 1 / 2, 0, 0], [0, 0, 1 / 2, 0], [1 / 2, 0, 0, 0]]
    )
    res = partial_transpose(rho)
    np.testing.assert_equal(np.allclose(res, expected_res), True)


def test_partial_transpose_non_square_matrix():
    """Matrix must be square."""
    with np.testing.assert_raises(ValueError):
        test_input_mat = np.array([[1, 2, 3, 4], [5, 6, 7, 8], [13, 14, 15, 16]])
        partial_transpose(test_input_mat)


def test_partial_transpose_non_square_matrix_2():
    """Matrix must be square."""
    with np.testing.assert_raises(ValueError):
        rho = np.array([[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12]])
        partial_transpose(rho, 2, [2])


def test_partial_transpose_cvxpy():
    """Test partial transpose on cvxpy objects."""
    x_var = cvxpy.Variable((4, 4), hermitian=True)
    x_pt = partial_transpose(x_var)
    np.testing.assert_equal(isinstance(x_pt, Vstack), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Test antisymmetric_projection."""
import numpy as np

from toqito.perms import antisymmetric_projection


def test_antisymmetric_projection_d_2_p_1():
    """Dimension is 2 and p is equal to 1."""
    res = antisymmetric_projection(2, 1).todense()
    expected_res = np.array([[1, 0], [0, 1]])

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_antisymmetric_projection_p_larger_than_d():
    """The `p` value is greater than the dimension `d`."""
    res = antisymmetric_projection(2, 3).todense()
    expected_res = np.zeros((8, 8))

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_antisymmetric_projection_2():
    """The dimension is 2."""
    res = antisymmetric_projection(2).todense()
    expected_res = np.array([[0, 0, 0, 0], [0, 0.5, -0.5, 0], [0, -0.5, 0.5, 0], [0, 0, 0, 0]])

    bool_mat = np.isclose(res, expected_res)
    np.testing.assert_equal(np.all(bool_mat), True)


def test_antisymmetric_projection_3_3_true():
    """The `dim` is 3, the `p` is 3, and `partial` is True."""
    res = antisymmetric_projection(3, 3, True).todense()
    np.testing.assert_equal(np.isclose(res[5].item(), -0.40824829), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Tests for is_trace_preserving."""
import numpy as np

from toqito.channel_props import is_trace_preserving
from toqito.channels import depolarizing


def test_is_trace_preserving_kraus_false():
    """Verify non-trace preserving channel as Kraus ops as False."""
    unitary_mat = np.array([[1, 1], [-1, 1]]) / np.sqrt(2)
    kraus_ops = [[np.identity(2), np.identity(2)], [unitary_mat, -unitary_mat]]

    np.testing.assert_equal(is_trace_preserving(kraus_ops), False)


def test_is_trace_preserving_choi_true():
    """Verify Choi matrix of the depolarizing map is trace preserving."""
    np.testing.assert_equal(is_trace_preserving(depolarizing(2)), True)


if __name__ == "__main__":
    np.testing.run_module_suite()



"""Calculate permutation sign."""
from typing import List, Union
from scipy import linalg

import numpy as np


def perm_sign(perm: Union[np.ndarray, List[int]]) -> float:
    """
    Compute the "sign" of a permutation [WikParPerm]_.

    The sign (either -1 or 1) of the permutation :code:`perm` is -1**`inv`, where `inv` is the
    number of inversions contained in :code:`perm`.

    Examples
    ==========

    For the following vector

    .. math::
        [1, 2, 3, 4]

    the permutation sign is positive as the number of elements in the vector are even. This can be
    performed in :code:`toqito` as follows.

    >>> from toqito.perms import perm_sign
    >>> perm_sign([1, 2, 3, 4])
    1

    For the following vector

    .. math::
        [1, 2, 3, 4, 5]

    the permutation sign is negative as the number of elements in the vector are odd. This can be
    performed in :code:`toqito` as follows.

    >>> from toqito.perms import perm_sign
    >>> perm_sign([1, 2, 4, 3, 5])
    -1

    References
    ==========
    .. [WikParPerm] Wikipedia: Parity of a permutation
        https://en.wikipedia.org/wiki/Parity_of_a_permutation

    :param perm: The permutation vector to be checked.
    :return: The value 1 if the permutation is of even length and the value of
             -1 if the permutation is of odd length.
    """
    eye = np.eye(len(perm))
    return linalg.det(eye[:, np.array(perm) - 1])



"""Isotropic state."""
import numpy as np

from toqito.states import max_entangled


def isotropic(dim: int, alpha: float) -> np.ndarray:
    r"""
    Produce a isotropic state [HH99]_.

    Returns the isotropic state with parameter :code:`alpha` acting on
    (:code:`dim`-by-:code:`dim`)-dimensional space. The isotropic state has the following form

    .. math::
        \begin{equation}
            \rho_{\alpha} = \frac{1 - \alpha}{d^2} \mathbb{I} \otimes
            \mathbb{I} + \alpha |\psi_+ \rangle \langle \psi_+ | \in
            \mathbb{C}^d \otimes \mathbb{C}^2
        \end{equation}

    where :math:`|\psi_+ \rangle = \frac{1}{\sqrt{d}} \sum_j |j \rangle \otimes |j \rangle` is the
    maximally entangled state.

    Examples
    ==========

    To generate the isotropic state with parameter :math:`\alpha=1/2`, we can make the following
    call to :code:`toqito` as

    >>> from toqito.states import isotropic
    >>> isotropic(3, 1 / 2)
    [[0.22222222, 0.        , 0.        , 0.        , 0.16666667,
      0.        , 0.        , 0.        , 0.16666667],
     [0.        , 0.05555556, 0.        , 0.        , 0.        ,
      0.        , 0.        , 0.        , 0.        ],
     [0.        , 0.        , 0.05555556, 0.        , 0.        ,
      0.        , 0.        , 0.        , 0.        ],
     [0.        , 0.        , 0.        , 0.05555556, 0.        ,
      0.        , 0.        , 0.        , 0.        ],
     [0.16666667, 0.        , 0.        , 0.        , 0.22222222,
      0.        , 0.        , 0.        , 0.16666667],
     [0.        , 0.        , 0.        , 0.        , 0.        ,
      0.05555556, 0.        , 0.        , 0.        ],
     [0.        , 0.        , 0.        , 0.        , 0.        ,
      0.        , 0.05555556, 0.        , 0.        ],
     [0.        , 0.        , 0.        , 0.        , 0.        ,
      0.        , 0.        , 0.05555556, 0.        ],
     [0.16666667, 0.        , 0.        , 0.        , 0.16666667,
      0.        , 0.        , 0.        , 0.22222222]]

    References
    ==========
    .. [HH99] Horodecki, Michał, and Paweł Horodecki.
        "Reduction criterion of separability and limits for a class of
        distillation protocols." Physical Review A 59.6 (1999): 4206.

    :param dim: The local dimension.
    :param alpha: The parameter of the isotropic state.
    :return: Isotropic state of dimension :code:`dim`.
    """
    psi = max_entangled(dim, False, False)
    return (1 - alpha) * np.identity(dim ** 2) / dim ** 2 + alpha * psi * psi.conj().T / dim



"""The partial trace."""
from typing import List, Union

import numpy as np

from cvxpy.expressions.expression import Expression
from cvxpy.expressions.variable import Variable
from toqito.perms import permute_systems
from toqito.helper import expr_as_np_array, np_array_as_expr


def partial_trace(
    input_mat: Union[np.ndarray, Variable],
    sys: Union[int, List[int]] = 2,
    dim: Union[int, List[int]] = None,
) -> Union[np.ndarray, Expression]:
    r"""
    Compute the partial trace of a matrix [WikPtrace]_.

    The *partial trace* is defined as

    .. math::
        \left( \text{Tr} \otimes \mathbb{I}_{\mathcal{Y}} \right)
        \left(X \otimes Y \right) = \text{Tr}(X)Y

    where :math:`X \in \text{L}(\mathcal{X})` and :math:`Y \in \text{L}(\mathcal{Y})` are linear
    operators over complex Euclidean spaces :math:`\mathcal{X}` and :math:`\mathcal{Y}`.

    Gives the partial trace of the matrix X, where the dimensions of the (possibly more than 2)
    subsystems are given by the vector :code:`dim` and the subsystems to take the trace on are
    given by the scalar or vector :code:`sys`.

    Examples
    ==========

    Consider the following matrix

    .. math::
        X = \begin{pmatrix}
                1 & 2 & 3 & 4 \\
                5 & 6 & 7 & 8 \\
                9 & 10 & 11 & 12 \\
                13 & 14 & 15 & 16
            \end{pmatrix}.

    Taking the partial trace over the second subsystem of :math:`X` yields the following matrix

    .. math::
        X_{pt, 2} = \begin{pmatrix}
                    7 & 11 \\
                    23 & 27
                 \end{pmatrix}

    By default, the partial trace function in :code:`toqito` takes the trace of the second
    subsystem.

    >>> from toqito.channels import partial_trace
    >>> import numpy as np
    >>> test_input_mat = np.array(
    >>>     [[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]]
    >>> )
    >>> partial_trace(test_input_mat)
    [[ 7, 11],
     [23, 27]]

    By specifying the :code:`sys = 1` argument, we can perform the partial trace over the first
    subsystem (instead of the default second subsystem as done above). Performing the partial
    trace over the first subsystem yields the following matrix

    .. math::
        X_{pt, 1} = \begin{pmatrix}
                        12 & 14 \\
                        20 & 22
                    \end{pmatrix}

    >>> from toqito.channels import partial_trace
    >>> import numpy as np
    >>> test_input_mat = np.array(
    >>>     [[1, 2, 3, 4], [5, 6, 7, 8], [9, 10, 11, 12], [13, 14, 15, 16]]
    >>> )
    >>> partial_trace(test_input_mat, 1)
    [[12, 14],
     [20, 22]]

    We can also specify both dimension and system size as :code:`list` arguments. Consider the
    following :math:`16`-by-:math:`16` matrix.

    >>> from toqito.channels import partial_trace
    >>> import numpy as np
    >>> test_input_mat = np.arange(1, 257).reshape(16, 16)
    >>> test_input_mat
    [[  1   2   3   4   5   6   7   8   9  10  11  12  13  14  15  16]
     [ 17  18  19  20  21  22  23  24  25  26  27  28  29  30  31  32]
     [ 33  34  35  36  37  38  39  40  41  42  43  44  45  46  47  48]
     [ 49  50  51  52  53  54  55  56  57  58  59  60  61  62  63  64]
     [ 65  66  67  68  69  70  71  72  73  74  75  76  77  78  79  80]
     [ 81  82  83  84  85  86  87  88  89  90  91  92  93  94  95  96]
     [ 97  98  99 100 101 102 103 104 105 106 107 108 109 110 111 112]
     [113 114 115 116 117 118 119 120 121 122 123 124 125 126 127 128]
     [129 130 131 132 133 134 135 136 137 138 139 140 141 142 143 144]
     [145 146 147 148 149 150 151 152 153 154 155 156 157 158 159 160]
     [161 162 163 164 165 166 167 168 169 170 171 172 173 174 175 176]
     [177 178 179 180 181 182 183 184 185 186 187 188 189 190 191 192]
     [193 194 195 196 197 198 199 200 201 202 203 204 205 206 207 208]
     [209 210 211 212 213 214 215 216 217 218 219 220 221 222 223 224]
     [225 226 227 228 229 230 231 232 233 234 235 236 237 238 239 240]
     [241 242 243 244 245 246 247 248 249 250 251 252 253 254 255 256]]

    We can take the partial trace on the first and third subsystems and assume that the size of
    each of the 4 systems is of dimension 2.

    >>> from toqito.channels import partial_trace
    >>> import numpy as np
    >>> partial_trace(test_input_mat, [1, 3], [2, 2, 2, 2])
    [[344, 348, 360, 364],
     [408, 412, 424, 428],
     [600, 604, 616, 620],
     [664, 668, 680, 684]])

    References
    ==========
    .. [WikPtrace] Wikipedia: Partial trace
        https://en.wikipedia.org/wiki/Partial_trace

    :param input_mat: A square matrix.
    :param sys: Scalar or vector specifying the size of the subsystems.
    :param dim: Dimension of the subsystems. If :code:`None`, all dimensions are assumed to be
                equal.
    :return: The partial trace of matrix :code:`input_mat`.
    """
    # If the input matrix is a CVX variable for an SDP, we convert it to a numpy array,
    # perform the partial trace, and convert it back to a CVX variable.
    if isinstance(input_mat, Variable):
        rho_np = expr_as_np_array(input_mat)
        traced_rho = partial_trace(rho_np, sys, dim)
        traced_rho = np_array_as_expr(traced_rho)
        return traced_rho

    if dim is None:
        dim = np.array([np.round(np.sqrt(len(input_mat)))])
    if isinstance(dim, int):
        dim = np.array([dim])
    if isinstance(dim, list):
        dim = np.array(dim)

    num_sys = len(dim)

    # Allow the user to enter a single number for dim.
    if num_sys == 1:
        dim = np.array([dim[0], len(input_mat) / dim[0]])
        if np.abs(dim[1] - np.round(dim[1])) >= 2 * len(input_mat) * np.finfo(float).eps:
            raise ValueError(
                "Invalid: If `dim` is a scalar, `dim` must evenly " "divide `len(input_mat)`."
            )
        dim[1] = np.round(dim[1])
        num_sys = 2

    prod_dim = np.prod(dim)
    if isinstance(sys, list):
        if len(sys) == 1:
            prod_dim_sys = np.prod(dim[0])
        else:
            prod_dim_sys = 1
            for idx in sys:
                prod_dim_sys *= dim[idx - 1]
    elif isinstance(sys, int):
        prod_dim_sys = np.prod(dim[sys - 1])
    else:
        raise ValueError(
            "Invalid: The variable `sys` must either be of type " "int or of a list of ints."
        )

    sub_prod = prod_dim / prod_dim_sys
    sub_sys_vec = prod_dim * np.ones(int(sub_prod)) / sub_prod

    if isinstance(sys, int):
        sys = [sys]
    set_diff = list(set(list(range(1, num_sys + 1))) - set(sys))

    perm = set_diff
    perm.extend(sys)

    a_mat = permute_systems(input_mat, perm, dim)

    ret_mat = np.reshape(
        a_mat,
        [int(sub_sys_vec[0]), int(sub_prod), int(sub_sys_vec[0]), int(sub_prod)],
        order="F",
    )
    permuted_mat = ret_mat.transpose((1, 3, 0, 2))

    permuted_reshaped_mat = np.reshape(
        permuted_mat,
        [int(sub_prod), int(sub_prod), int(sub_sys_vec[0] ** 2)],
        order="F",
    )

    pt_mat = permuted_reshaped_mat[
        :, :, list(range(0, int(sub_sys_vec[0] ** 2), int(sub_sys_vec[0] + 1)))
    ]
    pt_mat = np.sum(pt_mat, axis=2)

    return pt_mat
